from typing import Union
from abc import ABC, abstractmethod
import numpy as np
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist
from numba import njit
import MDAnalysis as mda

import chilife.RotamerEnsemble as re
import chilife.dRotamerEnsemble as dre
from .MolSys import MolSys


@njit(cache=True)
def get_lj_energy(r, rmin, eps, forgive=1, cap=10, rmax=10):
    """Return a vector with the energy values for the flat bottom lenard-jones potential from a set of atom pairs with
    distance `r`, rmin values of `rmin` and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.
    cap : flaot
        Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
        another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
        the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.
    rmax : float
        Maximum distance to consider for potential calculation. Any atom pairs with r > rmax will be set to 0.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.

    """
    r = np.atleast_2d(r)
    lj_energy = np.zeros_like(r)
    rmin_lower = forgive * rmin

    # Piecewise function for flat lj potential near rmin
    for j, rotamer in enumerate(r):
        for i in range(len(rotamer)):
            if rotamer[i] < rmin_lower[i]:
                lj = rmin_lower[i] / rotamer[i]
                lj = lj * lj * lj
                lj = lj * lj
                lj_energy[j, i] = np.minimum(eps[i] * (lj**2 - 2 * lj), cap * eps[i])

            elif rmin_lower[i] <= rotamer[i] < rmin[i]:
                lj_energy[j, i] = -eps[i]
            elif rotamer[i] < rmax:
                lj = rmin[i] / rotamer[i]
                lj = lj * lj * lj
                lj = lj * lj
                lj_energy[j, i] = eps[i] * (lj**2 - 2 * lj)

    return lj_energy


@njit(cache=True)
def get_lj_scwrl(r, rmin, eps, forgive=1):
    """Calculate a scwrl-like lenard-jones potential from a set of atom pairs with distance `r`, rmin values of `rmin`
    and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.
    """
    r = np.atleast_2d(r)
    lj_energy = np.empty_like(r)
    rmin_lower = rmin * forgive
    for j, rotamer in enumerate(r):
        # Piecewise function for flat lj potential near rmin
        for i in range(len(rotamer)):
            rat = rotamer[i] / (rmin_lower[i] / 1.12246204831)
            if rat < 0.8254:
                lj_energy[j, i] = 10 * eps[i]
            elif rat <= 1:
                lj_energy[j, i] = 57.273 * (1 - rat) * eps[i]
            elif rat < 10 / 9:
                lj_energy[j, i] = eps[i] * (10 - 9 * rat) ** (57.273 / (9 * eps[i])) - eps[i]
            elif rat < 4 / 3:
                lj_energy[j, i] = (eps[i] / 4) * (9 * rat - 10) ** 2 - eps[i]
            else:
                lj_energy[j, i] = 0

    return lj_energy


@njit(cache=True)
def get_lj_rep(r, rmin, eps, forgive=0.9, cap=10):
    """Calculate only the repulsive terms of the lenard-jones potential from a set of atom pairs with distance `r`,
    rmin values of `rmin` and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.
    cap : flaot
        Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
        another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
        the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.
    """
    r = np.atleast_2d(r)
    lj_energy = np.empty_like(r)
    rmin_lower = forgive * rmin

    # Piecewise function for flat lj potential near rmin
    for i, rotamer in enumerate(r):
        for j in range(len(rotamer)):
            lj = rmin_lower[j] / rotamer[j]
            lj = lj * lj * lj
            lj = lj * lj
            lj_energy[i, j] = np.minimum(eps[j] * lj**2, cap * eps[j])

    return lj_energy

@njit(cache=True)
def get_lj_attr(r, rmin, eps, forgive=0.9, floor=-2):
    """Calculate only the attractive terms of the lenard-jones potential from a set of atom pairs with distance `r`,
      rmin values of `rmin` and epsilon values of `eps`.

      Parameters
      ----------
      r : numpy.ndarray
          Vector of inter-atomic distances between non-bonded atoms of a system.
      rmin : numpy.ndarray
          Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
      eps : numpy.ndarray
          Vector of epsilon parameters corresponding to atom pairs of r
      forgive : numpy.ndarray
          The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
          and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
          value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
          in situations that would otherwise be resolved with minor atomic displacements.
      cap : flaot
          Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
          another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
          the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.

      Returns
      -------
      lj_energy: numpy.ndarray
          Vector of atom pair energies calculated using the modified lj potential function.
      """
    r = np.atleast_2d([r])
    lj_energy = np.empty_like(r)
    rmin_lower = forgive * rmin
    for j, rotamer in enumerate(r):
        # Piecewise function for flat lj potential near rmin
        for i in range(len(rotamer)):
            lj = rmin_lower[i] / rotamer[i]
            lj = lj * lj * lj
            lj = lj * lj
            lj_energy[j, i] = np.maximum(-2 * eps[i] * lj, eps[i] * floor)

    return lj_energy


def reweight_rotamers(energies, temp, weights):
    """Adjust rotamer population weights based on external energies (from clash evaluations).

    Parameters
    ----------
    energies : numpy.ndarray
        Array of external energies, in kcal/mol
    temp : scalar
        Temperature, in kelvin
    weights : numpy.ndarray
        Current weights of rotamers
    return_partition : bool
        If True, return the value of the partition function an additional argument; see below

    Returns
    -------
    new_weights : numpy.ndarray
        Adjusted weights
    partition : float (optional)
        The partition function relative to the free label. A small partition function suggests the interactions with
        neighboring atoms are unfavorable while a large partition function suggests the opposite.
    """

    probabilities = np.exp(-energies / (GAS_CONST * temp))
    p = probabilities * weights
    p_sum = np.sum(p)
    new_weights = p/p_sum
    partition = p_sum / weights.sum()

    return new_weights, partition


def join_geom(a, b, flat=False):
    """ Function to join Lennard-Jones parameters (``rmin`` or ``eps``) using their geometric mean. Parameters can be
    joined in two different ways (see keyword argument ``flat``)

    Parameters
    ----------
    a : numpy.ndarray
        Single atom parameters of atoms of group a
    b : numpy.ndarray
        Single atom parameters of atoms of group b
    flat : bool
        Only join parameters i of group a and j of group b if i=j. If false all combinations of i,j are computed.

    Returns
    -------
    param : numpy.ndarray
        The joined parameters of group a and b
    """

    if flat:
        return np.sqrt(a * b)
    else:
        return np.sqrt(np.outer(a, b))


def join_arith(a, b, flat=False):
    """ Function to join Lennard-Jones parameters (``rmin`` or ``eps``) using their arithmatic mean. Parameters can be
    joined in two different ways (see keyword argument ``flat``)

    Parameters
    ----------
    a : numpy.ndarray
        Single atom parameters of atoms of group a
    b : numpy.ndarray
        Single atom parameters of atoms of group b
    flat : bool
        Only join parameters i of group a and j of group b if i=j. If false all combinations of i,j are computed.

    Returns
    -------
    param : numpy.ndarray
        The joined parameters of group a and b
    """
    if flat:
        return a + b
    else:
        return np.add.outer(a, b)


rmin_charmm = {
     "C": 2.0000,
     "H": 1.27000,
     "N": 1.8500,
     "O": 1.7000,
     "S": 2.0000,
    "Se": 2.0000,  # Default Selenium to Sulfur
    "Br": 1.9800,
    "Cu": 0.7251,  # ion 2+
     'B': 2.5500,
     'F': 1.6300,
    'Na': 1.3638,  # ion
    'Mg': 1.1850,
     'P': 2.1500,
    'Cl': 1.9100,
     'K': 1.7638,  # ion
    'Ca': 1.3670,  # ion
    'Fe': 0.6500,  # HEM
    'Ni': 1.2760,
    'Zn': 1.0900,  # ion
    'Gd': 1.5050,  # ion 3+
    'Au': 0.6510,
    "join_protocol": join_arith,
}  # Bromobenzene from Gutiérrez et al. 2016

eps_charmm = {
    "C": -0.110,
    "H": -0.022,
    "N": -0.200,
    "O": -0.120,
    "S": -0.450,
    "Se": -0.450,
    "Br": -0.320,
    "Cu": -0.1505,
     'B': -0.0380,
     'F': -0.105,
    'Na': -0.0469,
    'Mg': -0.0150,
     'P': -0.5850,
    'Cl': -0.3430,
     'K': -0.0870,
    'Ca': -0.120,
    'Fe': 0.000,
    'Ni': -5.65,
    'Zn': -0.250,
    'Gd': -0.1723,
    'Au': -0.1330,
    "join_protocol": join_geom,
}  # Bromobenzene from Gutiérrez et al. 2016

rmin_uff = {
    "C": 3.851,
    "H": 2.886,
    "N": 3.660,
    "O": 3.500,
    "S": 4.035,
    "Se": 4.035,  # Default Selenium to Sulfur
    "Br": 4.189,
    "Na": 2.983,
    "Mg": 3.021,
    "P" : 4.147,
    "Cl": 3.947,
    "K" : 3.812,
    "Ca": 3.399,
    "Mn": 2.961,
    "Fe": 2.912,
    "Co": 2.872,
    "Ni": 2.834,
    "Cu": 3.495,
    "Gd": 3.368,
    "join_protocol": join_geom,
}

eps_uff = {
    "C": -0.105,
    "H": -0.044,
    "N": -0.069,
    "O": -0.060,
    "S": -0.274,
    "Se": -0.274,  # Default Selenium to Sulfur
    "Br": -0.251,
    "Na": -0.030,
    "Mg": -0.111,
    "P" : -0.305,
    "Cl": -0.227,
    "K" : -0.035,
    "Ca": -0.238,
    "Mn": -0.013,
    "Fe": -0.013,
    "Co": -0.014,
    "Ni": -0.015,
    "Cu": -0.005,
    "Gd": -0.009,
    "join_protocol": join_geom,
}

for dictionary in rmin_charmm, eps_charmm, rmin_uff, eps_uff:
    tdict = {**{key.upper(): val for key, val in dictionary.items()},
             **{key.lower(): val for key, val in dictionary.items()}}
    dictionary.update(tdict)

lj_params = {"uff": [rmin_uff, eps_uff], "charmm": [rmin_charmm, eps_charmm]}

class EnergyFunc:

    @abstractmethod
    def __call__(self, system):
        pass

    @abstractmethod
    def forces(self, system):
        pass

    @abstractmethod
    def prepare_system(self, system):
        pass

class ljEnergyFunc(EnergyFunc):
    def __init__(self, functional=None, params=None, extra_params=None, **kwargs):
        functional = get_lj_rep if functional is None else functional
        params = 'charmm' if params is None else params

        self.functional = functional
        self.rmax = kwargs.get("rmax", 10)
        self.forgive = kwargs.pop("forgive", 1)
        self.kwargs = kwargs


        if isinstance(params, dict):
            self._rmin_func, self._eps_func = params['rmin'], params['eps']
        elif params not in lj_params:
            raise RuntimeError(f'`{params}` is not a recognized forcefield.')
        else:
            self.name = params
            self._rmin_func, self._eps_func = lj_params[params]

        if extra_params is not None:
            self._rmin_func.update(extra_params['rmin'])
            self._eps_func.update(extra_params['eps'])


        self.join_rmin = self._rmin_func["join_protocol"]
        self.join_eps = self._eps_func["join_protocol"]

        self._system_hash = {}

    def get_lj_rmin(self, atypes):
        return np.array([self._rmin_func[a] for a in atypes])

    def get_lj_eps(self, atypes):
        return np.array([self._eps_func[a] for a in atypes])

    def prepare_system(self, system):

        if isinstance(system, (re.RotamerEnsemble, dre.dRotamerEnsemble)):
            # Prepare internal
            a, b = system.aidx, system.bidx
            a_eps = self.get_lj_eps(system.atom_types[a])
            a_radii = self.get_lj_rmin(system.atom_types[a])
            b_eps = self.get_lj_eps(system.atom_types[b])
            b_radii = self.get_lj_rmin(system.atom_types[b])

            system.irmin_ij = self.join_rmin(a_radii * self.forgive, b_radii * self.forgive, flat=True)
            system.ieps_ij = self.join_eps(a_eps, b_eps, flat=True)


            # Prepare external
            if system.protein is not None:
                environment_atypes = system.protein.atoms.types[system.protein_clash_idx]

                system.rmin2 = self.get_lj_rmin(system.atom_types[system.side_chain_idx])
                system.eps = self.get_lj_eps(system.atom_types[system.side_chain_idx])

                protein_lj_radii = self.get_lj_rmin(environment_atypes)
                protein_lj_eps = self.get_lj_eps(environment_atypes)

                system.ermin_ij = self.join_rmin(system.rmin2 * self.forgive, protein_lj_radii * self.forgive).reshape(-1)
                system.eeps_ij = self.join_eps(system.eps, protein_lj_eps).reshape((-1))

        elif isinstance(system, (mda.Universe, mda.AtomGroup, MolSys)):

            bonds = {(a, b) for a, b in system.atoms.bonds.indices}
            tree = cKDTree(system.atoms.positions)
            pairs = tree.query_pairs(self.rmax)
            pairs = pairs - bonds
            pairs = np.array(list(pairs))

            lj_radii_1 = self.get_lj_rmin(system.atoms.types[pairs[:, 0]])
            lj_radii_2 = self.get_lj_rmin(system.atoms.types[pairs[:, 1]])

            lj_eps_1 = self.get_lj_eps(system.atoms.types[pairs[:, 0]])
            lj_eps_2 = self.get_lj_eps(system.atoms.types[pairs[:, 1]])

            rmin_ij = self.join_rmin(lj_radii_1 * self.forgive, lj_radii_2*self.forgive, flat=True)
            eps_ij = self.join_eps(lj_eps_1, lj_eps_2, flat=True)

            if isinstance(system, MolSys):
                system.rmin_ij = rmin_ij
                system.eps_ij = eps_ij
                system.pairs = pairs
            else:
                self._system_hash[system] = rmin_ij, eps_ij, pairs

    def _get_functional_input(self, system, internal):
        if isinstance(system, (re.RotamerEnsemble, dre.dRotamerEnsemble)):
            if internal:
                rmin, eps = system.irmin_ij, system.ieps_ij
                r = np.linalg.norm(system.coords[:, system.aidx] - system.coords[:, system.bidx], axis=-1)
                shape = len(system), len(system.aidx)
            else:
                rmin, eps = system.ermin_ij, system.eeps_ij
                c1 = system.coords[:, system.side_chain_idx].reshape(-1, 3)
                c2 = system.protein_tree.data[system.protein_clash_idx]
                r = cdist(c1, c2).reshape(len(system), -1)
                shape = len(system), len(system.side_chain_idx)

        else:
            if isinstance(system, (mda.Universe, mda.AtomGroup)):
                rmin, eps, pairs = self._system_hash[system]
            else:
                rmin, eps, pairs = system.rmin_ij, system.eps_ij, system.pairs

            r = np.linalg.norm(system.positions[pairs[:,0]] - system.positions[pairs[:,1]], axis=-1)
            shape = len(system), len(pairs)

        return r, rmin, eps, shape

    def __call__(self, system, internal = False, **kwargs):
        tkwargs = {k: v for k, v in self.kwargs.items()}
        tkwargs.update(kwargs)
        r, rmin, eps, shape = self._get_functional_input(system, internal)
        E = self.functional(r, rmin, eps, **tkwargs)

        E = E.reshape(*shape, -1)
        if isinstance(system, (re.RotamerEnsemble, dre.dRotamerEnsemble)):
            system.atom_energies = E.sum(axis=2)

        return E.sum(axis=(1, 2))

KCAL2J = 4.184e3  # conversion factor form kcal to J (exact)
BOLTZ_CONST = 1.380649e-23  # Boltzmann constant, J K^-1 (exact)
AVOGADRO_CONST = 6.02214076e23  # Avogadro constant, mol^-1 (exact)
GAS_CONST = BOLTZ_CONST*AVOGADRO_CONST/KCAL2J  # molar gas constant, kcal mol^1 K^-1



import igraph as ig
from copy import deepcopy
from itertools import combinations
import logging
import warnings
import inspect
from functools import partial
from pathlib import Path

import numpy as np
from scipy.spatial import cKDTree
import scipy.optimize as opt
import MDAnalysis as mda

import chilife.io as io
import chilife.RotamerEnsemble as re
import chilife.scoring as scoring

from chilife.protein_utils import make_mda_uni
from .MolSysIC import MolSysIC

class dRotamerEnsemble:
    """Create new dRotamerEnsemble object.

        Parameters
        ----------
        res : string
            3-character name of desired residue, e.g. RXA.
        site : tuple[int, int]
            MolSys residue numbers to attach the bifunctional library to.
        protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
            Object containing all protein information (coords, atom types, etc.)
        chain : str
            MolSys chain identifier to attach the bifunctional ensemble to.
        rotlib : str
            Rotamer library to use for constructing the RotamerEnsemble
        **kwargs : dict
            restraint_weight: float
                Force constant (kcal/mol/A^2) for calculating energetic penalty of restraint satisfaction, i.e. the
                alignment of the overlapping atoms of the two mono-functional subunits of the bifunctional label.
            torsion_weight: float
                Force constant (kcal/mol/radian^2) for calculating energetic penalty of the deviation from rotamer
                starting dihedral angles.
            minimize: bool
                Switch to turn on/off minimization. During minimization each rotamer is optimized in dihedral space
                with respect to alignment of the "cap" atoms of the two mono-functional subunits, internal clashes
                and deviation from the starting conformation in dihedral space.
            min_method: str
                Name of the minimization algorithm to use. All ``scipy.optimize.minimize`` algorithms are available
                and include: ‘Nelder-Mead’, ‘Powell’, ‘CG’, ‘BFGS’, ‘Newton-CG’, ‘L-BFGS-B’, ‘TNC’, ‘COBYLA’, ‘SLSQP’,
                ‘trust-constr’, ‘dogleg’, ‘trust-ncg’, ‘trust-exact’, ‘trust-krylov’, and custom.
            exclude_nb_interactions: int:
                When calculating internal clashes, ignore 1-``exclude_nb_interactions`` interactions and below. Defaults
                to ignore 1-3 interactions, i.e. atoms that are connected by 2 bonds or fewer will not have a steric
                effect on each other.
            eval_clash : bool
                Switch to turn clash evaluation on (True) and off (False).
            forcefield: str
                Name of the forcefield you wish to use to parameterize atoms for the energy function. Currently,
                supports `charmm` and `uff`
            energy_func : callable
               Python function or callable object that takes a protein or a RotamerEnsemble object as input and
               returns an energy value (kcal/mol) for each rotamer in the ensemble. See also
               :mod:`Scoring <chiLife.scoring>`. Defaults to a capped lennard-jones potentail.
            temp : float
               Temperature to use when running ``energy_func``. Only used if ``energy_func`` accepts a temperature
               argument  Defaults to 298 K.
            clash_radius : float
               Cutoff distance (angstroms) for inclusion of atoms in clash evaluations. This distance is measured from
               ``clash_ori`` Defaults to the longest distance between any two atoms in the rotamer ensemble plus 5
               angstroms.
            clash_ori : str
               Atom selection to use as the origin when finding atoms within the ``clash_radius``. Defaults to 'cen',
               the centroid of the rotamer ensemble heavy atoms.
            protein_tree : Scipy.spatial.cKDTree
               KDTree of atom positions for fast distance calculations and neighbor detection. Defaults to None
            trim: bool
                When true, the lowest `trim_tol` fraction of rotamers in the ensemble will be removed.
            trim_tol: float
                Tolerance for trimming rotamers from the ensemble. trim_tol=0.005 means the bottom 0.5% of rotamers
                will be removed.
            alignment_method : str
               Method to use when attaching or aligning the rotamer ensemble backbone with the protein backbone.
               Defaults to ``'bisect'`` which aligns the CA atom, the vectors bisecting the N-CA-C angle and the
               N-CA-C plane.
            dihedral_sigmas : float, numpy.ndarray
               Standard deviations of dihedral angles (degrees) for off rotamer sampling. Can be a single number for
               isotropic sampling, a vector to define each dihedral individually or a matrix to define a value for
               each rotamer and each dihedral. Setting this value to np.inf will force uniform (accessible volume)
               sampling. Defaults to 35 degrees.
            weighted_sampling : bool
               Determines whether the rotamer ensemble is sampled uniformly or based off of their intrinsic weights.
               Defaults to False.
            use_H : bool
               Determines if hydrogen atoms are used or not. Defaults to False.

        Attributes
        ----------
        res : string
            3-character name of desired residue, e.g. RXA.
        site1 : int
            MolSys residue number of the first attachment site.
        site2 : int
            MolSys residue number of the second attachment site.
        increment : int
            Number of residues  between two attachment sites.
        protein : MDAnalysis.Universe, MDAnalysis.AtomGroup, :class:`~MolSys`
            Object containing all protein information (coords, atom types, etc.)
        chain : str
            Chain identifier of the site the bifunctional ensemble is attached to.
        name : str
            Name of the ensemble. Usually include the native amino acid, the site number and the label that was attached
            changing the name will change the object name when saving in a PDB.
        RE1 : RotamerEnsemble
            Monofunctional ensemble subunit attached to the first site.
        RE2 : RotamerEnsemble
            Monofunctional ensemble subunit attached to the second site.
    """


    backbone_atoms = ["H", "N", "CA", "HA", "C", "O"]

    def __init__(self, res, sites, protein=None, chain=None, rotlib=None, **kwargs):
        self.res = res
        self.site1, self.icode1, self.site2, self.icode2 = proc_sites(sites)

        self.site = self.site1
        self.increment = self.site2 - self.site1
        self.kwargs = kwargs

        self.protein = protein
        self.chain = chain if chain is not None else self.guess_chain()

        self.input_kwargs = kwargs
        self.__dict__.update(dassign_defaults(kwargs))

        self.get_lib(rotlib)
        self.create_ensembles()

        self.RE1.backbone_to_site()
        self.RE2.backbone_to_site()

        self.cst_idx1 = np.where(self.RE1.atom_names[None, :] == self.csts[:, None])[1]
        self.cst_idx2 = np.where(self.RE2.atom_names[None, :] == self.csts[:, None])[1]

        _, idx1 = np.unique(self.cst_idx1, return_index=True)
        _, idx2 = np.unique(self.cst_idx2, return_index=True)

        self.cst_idx1 = self.cst_idx1[np.sort(idx1)]
        self.cst_idx2 = self.cst_idx2[np.sort(idx2)]

        for i in range(1, len(self.cst_idx2)):
            if self.RE2.atom_names[self.cst_idx2[i]] == self.RE2.atom_names[self.cst_idx2[i-1]]:
                self.cst_idx2[i - 1], self.cst_idx2[i] = self.cst_idx2[i], self.cst_idx2[i - 1]

        self.rl1mask = np.argwhere(~np.isin(self.RE1.atom_names, self.csts)).flatten()
        self.rl2mask = np.argwhere(~np.isin(self.RE2.atom_names, self.csts)).flatten()

        self.name = self.res
        if self.site1 is not None:
            self.name = f"{self.RE1.nataa}{self.site1}-{self.RE2.nataa}{self.site2}{self.res}"
        if self.chain is not None:
            self.name += f"_{self.chain}"

        self.selstr = f"resid {self.site1} {self.site2} and segid {self.chain} and not altloc B"

        self._graph = ig.Graph(edges=self.bonds)

        if self.clash_radius is None:
            self.clash_radius = np.linalg.norm(self.clash_ori - self.coords, axis=-1).max() + 5

        self.protein_setup()
        self.sub_labels = (self.RE1, self.RE2)

    @property
    def weights(self):
        """Array of the fractions of rotamer populations for each rotamer in the library."""
        return self.RE1.weights

    @weights.setter
    def weights(self, value):
        self.RE1.weights = value
        self.RE2.weights = value

    @property
    def coords(self):
        """The 3D cartesian coordinates of each atom of each rotamer in the library."""
        ovlp = (self.RE1.coords[:, self.cst_idx1] + self.RE2.coords[:, self.cst_idx2]) / 2
        return np.concatenate([self.RE1._coords[:, self.rl1mask], self.RE2._coords[:, self.rl2mask], ovlp], axis=1)

    @coords.setter
    def coords(self, value):
        if value.shape[1] != len(self.atom_names):
            raise ValueError(
                f"The provided coordinates do not match the number of atoms of this ensemble ({self.res})"
            )

        self.RE1._coords[:, self.rl1mask] = value[:, :len(self.rl1mask)]
        self.RE2._coords[:, self.rl2mask] = value[:, len(self.rl1mask):len(self.rl1mask) + len(self.rl2mask)]
        self.RE1._coords[:, self.cst_idx1] = value[:, len(self.rl1mask) + len(self.rl2mask):]
        self.RE2._coords[:, self.cst_idx2] = value[:, len(self.rl1mask) + len(self.rl2mask):]

    @property
    def _lib_coords(self):
        ovlp = (self.RE1._lib_coords[:, self.cst_idx1] + self.RE2._lib_coords[:, self.cst_idx2]) / 2
        return np.concatenate([self.RE1._lib_coords[:, self.rl1mask],
                               self.RE2._lib_coords[:, self.rl2mask], ovlp], axis=1)

    @property
    def atom_names(self):
        """The names of each atom in the rotamer"""
        return np.concatenate((self.RE1.atom_names[self.rl1mask],
                               self.RE2.atom_names[self.rl2mask],
                               self.RE1.atom_names[self.cst_idx1]))

    @property
    def atom_types(self):
        """The element or atom type of each atom in the rotamer."""
        return np.concatenate((self.RE1.atom_types[self.rl1mask],
                               self.RE2.atom_types[self.rl2mask],
                               self.RE1.atom_types[self.cst_idx1]))

    @property
    def dihedral_atoms(self):
        """Four atom sets defining each flexible dihedral of the side chain"""
        return np.concatenate([self.RE1.dihedral_atoms, self.RE2.dihedral_atoms])

    @property
    def dihedrals(self):
        """Dihedral angle values of each dihedral defined in :py:attr::`~dihedral_atoms` for each rotamer in the
        library"""
        return np.concatenate([self.RE1.dihedrals, self.RE2.dihedrals], axis=-1)

    @property
    def centroid(self):
        """The geometric center of all atoms of all rotamers in the rotamer library"""
        return self.coords.mean(axis=(0, 1))

    @property
    def clash_ori(self):
        """The origin used to determine if an external atom will be considered for clashes using the
        ``clash_radius`` property of the ensemble"""

        if isinstance(self._clash_ori_inp, (np.ndarray, list)):
            if len(self._clash_ori_inp) == 3:
                return self._clash_ori_inp

        elif isinstance(self._clash_ori_inp, str):
            if self._clash_ori_inp in ["cen", "centroid"]:
                return self.centroid

            elif (ori_name := self._clash_ori_inp.upper()) in self.atom_names:
                return np.squeeze(self.coords[0][ori_name == self.atom_names])

        else:
            raise ValueError(
                f"Unrecognized clash_ori option {self._clash_ori_inp}. Please specify a 3D vector, an "
                f"atom name or `centroid`"
            )

        return self._clash_ori

    @clash_ori.setter
    def clash_ori(self, inp):
        self._clash_ori_inp = inp

    @property
    def side_chain_idx(self):
        """Indices of the atoms that correspond to the side chain atoms (e.g. CB, CG, etc. and not N, CA, C)"""
        if not hasattr(self, '_side_chain_idx'):
            self._side_chain_idx = np.argwhere(
                np.isin(self.atom_names, dRotamerEnsemble.backbone_atoms, invert=True)
            ).flatten()

        return self._side_chain_idx

    @property
    def bonds(self):
        """Array of intra-label atom pairs indices that are covalently bonded."""
        if not hasattr(self, "_bonds"):
            bonds = []

            for bond in self.RE1.bonds:
                bndin = np.isin(bond, self.rl1mask)
                if np.all(bndin):
                    bonds.append(bond)
                elif np.any(bndin):
                    bonds.append([bond[0], np.argwhere(self.atom_names == self.RE1.atom_names[bond[1]]).flat[0]])
                else:
                    bonds.append([np.argwhere(self.atom_names == self.RE1.atom_names[bond[0]]).flat[0],
                                  np.argwhere(self.atom_names == self.RE1.atom_names[bond[1]]).flat[0]])

            for bond in self.RE2.bonds:
                bndin = np.isin(bond, self.rl2mask)
                if np.all(bndin):
                    bonds.append([b + len(self.rl1mask) for b in bond])
                elif not bndin[1]:
                    bonds.append([bond[0] + len(self.rl1mask),
                                  np.argwhere(self.atom_names == self.RE2.atom_names[bond[1]]).flat[0]])

            self._bonds = np.array(sorted(set(map(tuple, bonds))), dtype=int)

        return self._bonds

    @bonds.setter
    def bonds(self, inp):
        """
        Set of intra-label bonded pairs.
        Parameters
        ----------
        inp : ArrayLike
            List of atom ID pairs that are bonded
        """
        self._bonds = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._non_bonded = all_pairs - self._bonds

    @property
    def non_bonded(self):
        """ Array of indices of intra-label non-bonded atom pairs. Primarily used for internal clash evaluation when
        sampling the dihedral space"""

        if not hasattr(self, "_non_bonded"):
            pairs = {v.index: [path for path in self._graph.get_all_shortest_paths(v) if
                           len(path) <= (self._exclude_nb_interactions)] for v in self._graph.vs}
            pairs = {(a, c) for a in pairs for b in pairs[a] for c in b if a < c}
            all_pairs = set(combinations(range(len(self.atom_names)), 2))
            self._non_bonded = all_pairs - pairs

        return sorted(list(self._non_bonded))

    @non_bonded.setter
    def non_bonded(self, inp):
        """
        Create a set of non-bonded atom pair indices within a single side chain
        Parameters
        ----------
        inp : ArrayLike
            List of atom ID pairs that are not bonded
        """

        self._non_bonded = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._bonds = all_pairs - self._non_bonded

    def protein_setup(self):
        if isinstance(self.protein, (mda.AtomGroup, mda.Universe)):
            if not hasattr(self.protein.universe._topology, "altLocs"):
                self.protein.universe.add_TopologyAttr('altLocs', np.full(len(self.protein.universe.atoms), ""))
        if self.ignore_waters:
            self.protein = self.protein.select_atoms("not (byres name OH2 or resname HOH)")
        else:
            self.protein = self.protein.atoms

        clash_ignore_idx = self.protein.select_atoms(f"resid {self.site1} {self.site2} and segid {self.chain}").ix
        self.clash_ignore_idx = np.argwhere(np.isin(self.protein.ix, clash_ignore_idx)).flatten()
        self.resindex = self.protein.select_atoms(self.selstr).residues[0].resindex
        self.segindex = self.protein.select_atoms(self.selstr).residues[0].segindex

        if self.protein_tree is None:
            self.protein_tree = cKDTree(self.protein.atoms.positions)

        protein_clash_idx = self.protein_tree.query_ball_point(
            self.clash_ori, self.clash_radius
        )
        self.protein_clash_idx = [
            idx for idx in protein_clash_idx if idx not in self.clash_ignore_idx
        ]

        self.aidx, self.bidx = [list(x) for x in zip(*self.non_bonded)]
        if hasattr(self.energy_func, 'prepare_system'):
            self.energy_func.prepare_system(self)

        if self._minimize:
            self.minimize()

        if self.eval_clash:
            self.evaluate()

    def guess_chain(self):
        """
        Function to guess the chain based off of the attached protein and residue number.

        Returns
        -------
        chain : str
            The chain name
        """
        if self.protein is None:
            chain = "A"
        elif len(set(self.protein.segments.segids)) == 1:
            chain = self.protein.segments.segids[0]
        elif np.isin(self.protein.residues.resnums, self.site1).sum() == 0:
            raise ValueError(
                f"Residue {self.site1} is not present on the provided protein"
            )
        elif np.isin(self.protein.residues.resnums, self.site1).sum() == 1:
            chain = self.protein.select_atoms(f"resid {self.site1}").segids[0]
        else:
            raise ValueError(
                f"Residue {self.site1} is present on more than one chain. Please specify the desired chain"
            )
        return chain

    def get_lib(self, rotlib):
        """
        Specil function to get a rotamer library for a dRotamerEnsemble and apply all attributes to the object
        instance. ``rotlib`` can be a residue name, a file name, or a dictionary containing all  the standard entries
        of a rotamer library.

        Parameters
        ----------
        rotlib : Union[Path, str, dict]
            The name of the residue, the Path to the rotamer library file or a dictionary containing all entries of a
            rotamer library
        """

        # If given a dictionary use that as the rotlib
        if isinstance(rotlib, dict):
            if tuple(sorted(rotlib.keys())) != ('csts', 'libA', 'libB'):
                raise RuntimeError('Non-file dRotamerEnsemble rotlibs must be dictionaries consisting of exactle 3 '
                                   'entries: `csts`, `libA` and `libB`')
            self.csts = rotlib['csts']
            self.libA, self.libB = rotlib['libA'], rotlib['libB']
            self.kwargs["eval_clash"] = False
            return None

        rotlib = self.res if rotlib is None else rotlib
        if 'ip' not in rotlib:
            rotlib += f'ip{self.increment}'

        # Check if any exist
        rotlib_path = io.get_possible_rotlibs(rotlib, suffix='drotlib', extension='.zip')

        if rotlib_path is None:
            # Check if libraries exist but for different i+n
            rotlib_path = io.get_possible_rotlibs(rotlib.replace(f'ip{self.increment}', ''),
                                                       suffix='drotlib',
                                                       extension='.zip',
                                                       return_all=True)

            if rotlib_path is None:
                raise NameError(f'There is no rotamer library called {rotlib} in this directory or in chilife')

            else:
                warnings.warn(f'No rotamer library found for the given increment (ip{self.increment}) but rotlibs '
                              f'were found for other increments. chiLife will combine these rotlib to model '
                              f'this site1 pair and but they may not be accurate! Because there is no information '
                              f'about the relative weighting of different rotamer libraries all weights will be '
                              f'set to 1/len(rotlib)')

        if isinstance(rotlib_path, Path):
            libA, libB, csts = io.read_library(rotlib_path)

        elif isinstance(rotlib_path, list):

            cctA, cctB, ccsts = {}, {}, {}
            libA, libB, csts = io.read_library(rotlib_path[0])
            unis = []

            for lib in (libA, libB):
                names, types = lib['atom_names'], libA['atom_types']
                residxs = np.zeros(len(names), dtype=int)
                resnames, resids = np.array([self.res]), np.array([1])
                segidx = np.array([0])

                uni = make_mda_uni(names, types, resnames, residxs, resids, segidx)
                unis.append(uni)

            for p in rotlib_path:
                tlibA, tlibB, tcsts = io.read_library(p)
                for lib, tlib, cct, uni in zip((libA, libB), (tlibA, tlibB), (cctA, cctB), unis):

                    # Libraries must have the same atom order
                    if not np.all(np.isin(tlib['atom_names'], lib['atom_names'])) and \
                            np.all(tlib['dihedral_atoms'] == lib['dihedral_atoms']):
                        raise ValueError(f'Rotlibs {rotlib_path[0].stem} and {p.stem} are not compatable. You may'
                                         f'need to rename one of them.')

                    # Map coordinates
                    ixmap = [np.argwhere(tlib['atom_names'] == aname).flat[0] for aname in lib['atom_names']]
                    cct.setdefault('coords', []).append(tlib['coords'][:, ixmap])

                    # Create new internal coords if they are defined differently
                    lib_ic, tlib_ic = lib['internal_coords'], tlib['internal_coords']
                    if np.any(lib_ic.atom_names != tlib_ic.atom_names):
                        uni.load_new(cct['coords'][-1])
                        tlib_ic = MolSysIC.from_atoms(uni, lib['dihedral_atoms'], lib_ic.bonds)

                    tlib['zmats'] = tlib_ic.trajectory.coordinate_array
                    cct.setdefault('dihedrals', []).append(tlib['dihedrals'])
                    cct.setdefault('zmats', []).append(tlib['zmats'])

            for field in ('dihedrals', 'coords', 'zmats'):
                libA[field] = np.concatenate(cctA[field])
                libB[field] = np.concatenate(cctB[field])

            libA['internal_coords'].load_new(libA.pop('zmats'))
            libB['internal_coords'].load_new(libB.pop('zmats'))
            libA['weights'] = libB['weights'] = np.ones(len(libA['coords'])) / len(libA['coords'])

        self.csts = csts
        self.libA, self.libB = libA, libB
        self.kwargs["eval_clash"] = False

    def create_ensembles(self):
        """Creates monofunctional components of the bifunctional rotamer ensemble."""

        self.RE1 = re.RotamerEnsemble(self.res,
                                   self.site1,
                                   self.protein,
                                   self.chain,
                                   self.libA,
                                   **self.kwargs)

        self.RE2 = re.RotamerEnsemble(self.res,
                                   self.site2,
                                   self.protein,
                                   self.chain,
                                   self.libB,
                                   **self.kwargs)


    def save_pdb(self, name: str = None):
        """
        Save a PDB file of the ensemble

        Parameters
        ----------
        name: str
            Name of the PDB file
        """
        if name is None:
            name = self.name + ".pdb"
        if not name.endswith(".pdb"):
            name += ".pdb"

        io.save(name, self.RE1, self.RE2)

    def minimize(self, callback=None):
        """
        Minimize rotamers in dihedral space in the current context. Performed by default unless the ``minimize=False``
        keyword argument is used during construction. Note that the minimization method is controlled by the

        Parameters
        ----------
        callback: Callable
            A callable function to be passed as the ``scipy.optimize.minimize`` function. See the
            `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            for details.
        """

        scores = [self._min_one(i, ic1, ic2, callback=callback) for i, (ic1, ic2) in
                  enumerate(zip(self.RE1.internal_coords, self.RE2.internal_coords))]

        scores = np.asarray(scores)
        SSEs = np.linalg.norm(self.RE1.coords[:, self.cst_idx1] - self.RE2.coords[:, self.cst_idx2], axis=2).sum(axis=1)
        MSD = SSEs / len(self.csts)
        MSDmin = MSD.min()

        if MSDmin > 0.1:
            warnings.warn(f'The minimum MSD of the cap is {MSD.min()}, this may result in distorted label. '
                          f'Check that the structures make sense.')

        if MSDmin > 0.25:
            raise RuntimeError(f'chiLife was unable to connect residues {self.site1} and {self.site2} with {self.res}. '
                               f'Please double check that this is the intended labeling site. It is likely that these '
                               f'sites are too far apart.')
        self.cap_MSDs = MSD
        self.RE1.backbone_to_site()
        self.RE2.backbone_to_site()
        self.score_base = scores.min()
        scores -= scores.min()
        self.rotamer_scores = scores + self.score_base
        self.weights *= np.exp(-scores / (scoring.GAS_CONST * self.temp) / np.exp(-scores).sum())
        self.weights /= self.weights.sum()

    def _objective(self, dihedrals, ic1, ic2):
        """
        Objective function to optimize for each rotamer in the ensemble.

        Parameters
        ----------
        dihedrals: ArrayLike
            Dihedral values

        ic1, ic2: chiLife.MolSysIC
            Internal coordinates object for the two mono functional subunits of the bifunctional label.

        Returns
        -------
        score: float
            Rotamer energy score for the current conformation
        """

        ic1.set_dihedral(dihedrals[: len(self.RE1.dihedral_atoms)], 1, self.RE1.dihedral_atoms)
        coords1 = ic1.to_cartesian()[self.RE1.ic_mask]

        ic2.set_dihedral(dihedrals[-len(self.RE2.dihedral_atoms):], 1, self.RE2.dihedral_atoms)
        coords2 = ic2.to_cartesian()[self.RE2.ic_mask]

        diff = np.linalg.norm(coords1[self.cst_idx1] - coords2[self.cst_idx2], axis=1)
        ovlp = (coords1[self.cst_idx1] + coords2[self.cst_idx2]) / 2
        coords = np.concatenate([coords1[self.rl1mask], coords2[self.rl2mask], ovlp], axis=0)
        r = np.linalg.norm(coords[self.aidx] - coords[self.bidx], axis=1)

        # Faster to compute lj here
        lj = self.irmin_ij / r
        lj = lj * lj * lj
        lj = lj * lj

        # attractive forces are needed, otherwise this term will perpetually push atoms apart
        internal_energy = self.ieps_ij * (lj * lj - 2 * lj)
        score = (diff @ diff) * self.restraint_weight / len(diff) + internal_energy.sum()

        return score

    def _min_one(self, i, ic1, ic2, callback=None):
        """
        Helper function to use when dispatching minimization jobs or each rotamer.

        Parameters
        ----------
        i: int
            rotamer index
        ic1, ic2: chiLife.MolSysIC
            Internal coordinates object for the two mono functional subunits of the bifunctional label.
        callback: Callable
            A callable function to be passed as the ``scipy.optimize.minimize`` function. See the
            `scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            for details.

        Returns
        -------
        score: float
            Energy score of the minimized rotamer.
        """

        if callback is not None:
            if 'i' in inspect.signature(callback).parameters:
                callback = partial(callback, i=i)


        d0 = np.concatenate([ic1.get_dihedral(1, self.RE1.dihedral_atoms),
                             ic2.get_dihedral(1, self.RE2.dihedral_atoms)])

        lb = d0 - np.pi  # np.deg2rad(40)
        ub = d0 + np.pi  # np.deg2rad(40) #
        bounds = np.c_[lb, ub]
        xopt = opt.minimize(self._objective, x0=d0, args=(ic1, ic2),
                            bounds=bounds, method=self.min_method,
                            callback=callback)
        self.RE1._coords[i] = ic1.coords[self.RE1.H_mask]
        self.RE2._coords[i] = ic2.coords[self.RE2.H_mask]
        tors = d0 - xopt.x
        tors = np.arctan2(np.sin(tors), np.cos(tors))
        tors = np.sqrt(tors @ tors)

        return xopt.fun + tors * self.torsion_weight

    def trim_rotamers(self):
        """Remove low probability rotamers from the ensemble. All rotamers accounting for the population less than
        ``self.trim_tol`` will be removed."""

        arg_sort_weights = np.argsort(self.weights)[::-1]
        sorted_weights = self.weights[arg_sort_weights]
        cumulative_weights = np.cumsum(sorted_weights)
        cutoff = np.maximum(1, len(cumulative_weights[cumulative_weights < 1 - self.trim_tol]))
        keep_idx = arg_sort_weights[:cutoff]
        if hasattr(self, 'cap_MSDs'):
            self.cap_MSDs = self.cap_MSDs[keep_idx]
            self.rotamer_scores = self.rotamer_scores[keep_idx]

        if hasattr(self, 'rot_clash_energy'):
            self.rot_clash_energy = self.rot_clash_energy[keep_idx]
        self.RE1.trim_rotamers(keep_idx=keep_idx)
        self.RE2.trim_rotamers(keep_idx=keep_idx)

    def evaluate(self):
        """Place rotamer ensemble on protein site and recalculate rotamer weights."""
        # Calculate external energies
        energies = self.energy_func(self)
        self.rot_clash_energy = energies
        # Calculate total weights (combining internal and external)
        self.weights, self.partition = scoring.reweight_rotamers(energies, self.temp, self.weights)
        logging.info(f"Relative partition function: {self.partition:.3}")

        # Remove low-weight rotamers from ensemble
        if self._do_trim:
            self.trim_rotamers()

    def __len__(self):
        """Return the length of the rotamer library (number of rotamers)"""
        return len(self.RE1.coords)

    def copy(self):
        """Create a deep copy of the dRotamerEnsemble object"""
        new_copy = dRotamerEnsemble(self.res, (self.site1, self.site2), chain=self.chain,
                                    protein=self.protein,
                                    rotlib={'csts': self.csts, 'libA': self.libA, 'libB': self.libB},
                                    minimize=False,
                                    eval_clash=False)
        for item in self.__dict__:
            if isinstance(self.dict[item], np.ndarray):
                new_copy.__dict__[item] = self.__dict__[item].copy()

            elif item == 'RE1':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libA)

            elif item == 'RE2':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libB)

            elif item == 'protein':
                pass

            else:
                new_copy.__dict__[item] = deepcopy(self.__dict__[item])
        return new_copy

def proc_sites(sites):
    sites = sorted(sites)
    new_sites = []
    for site in sites:
        if isinstance(site, str):
            site_split = re.split('(\D+)',site)
            site = site_split[0]
            icode = site_split[1] if len(site_split) > 1 else ""
        else:
            icode = ""
        new_sites.append(site)
        new_sites.append(icode)

    return new_sites

def dassign_defaults(kwargs):
    """
    Helper function to assign default values to kwargs that have not been explicitly assigned by the user. Also
    checks to make sure that the user provided kwargs are real kwargs.
    Parameters
    ----------
    kwargs : dict
        Dictionary of user supplied keyword arguments.


    Returns
    -------
    kwargs : dict
        Dictionary of user supplied keyword arguments augmented with defaults for all kwargs the user did not supply.
    """

    # Make all string arguments lowercase
    for key, value in kwargs.items():
        if isinstance(value, str):
            kwargs[key] = value.lower()

    # Default parameters
    defaults = {
        "eval_clash": True,
        "temp": 298,
        "clash_radius": None,
        "protein_tree": None,
        "_clash_ori_inp": kwargs.pop("clash_ori", "cen"),

        "alignment_method": "bisect",
        "dihedral_sigmas": 25,

        "use_H": False,
        "_exclude_nb_interactions": kwargs.pop('exclude_nb_interactions', 3),

        "energy_func": scoring.ljEnergyFunc(scoring.get_lj_energy, 'charmm', forgive=0.95)
,
        "_minimize": kwargs.pop('minimize', True),
        "min_method": 'L-BFGS-B',
        "_do_trim": kwargs.pop('trim', True),
        "trim_tol": 0.005,

        "restraint_weight": kwargs.pop('restraint_weight', 222),
        "torsion_weight": 5,

        "ignore_waters": True,
    }

    # Overwrite defaults
    kwargs_filled = {**defaults, **kwargs}

    # Make sure there are no unused parameters
    if len(kwargs_filled) != len(defaults):
        raise TypeError(
            f"Got unexpected keyword argument(s): "
            f'{", ".join(key for key in kwargs if key not in defaults)}'
        )

    return kwargs_filled.items()


import logging
logging.getLogger("MDAnalysis").setLevel(logging.ERROR)

from .chilife import (distance_distribution, confidence_interval, create_library, create_dlibrary, add_rotlib_dir,
                      remove_rotlib_dir, add_library, remove_library, add_dihedral_def, add_to_defaults,
                      remove_from_defaults, list_available_rotlibs, rotlib_info, repack, add_to_toml)
from .globals import *
from .io import *
from .alignment_methods import parse_backbone, global_mx, local_mx
from .RotamerEnsemble import RotamerEnsemble
from .dRotamerEnsemble import dRotamerEnsemble
from .SpinLabel import SpinLabel
from .dSpinLabel import dSpinLabel
from .IntrinsicLabel import IntrinsicLabel
from .SpinLabelTraj import *
from .MolSys import *
from .MolSysIC import *
from .Topology import *
from .protein_utils import *
from .scoring import *


# RotamerEnsemble = RotamerEnsemble.RotamerEnsemble
# dRotamerEnsemble = dRotamerEnsemble.dRotamerEnsemble

# SpinLabel = SpinLabel.SpinLabel
# dSpinLabel = dSpinLabel.dSpinLabel

__version__ = '1.1.4'



import hashlib, os, pickle
from functools import partial
import numpy as np
import chilife
import pytest
import MDAnalysis as mda

ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
U = mda.Universe("test_data/1omp.pdb", in_memory=True)
exou = mda.Universe("test_data/3tu3.pdb", in_memory=True)
traj = mda.Universe('test_data/xlsavetraj.pdb', in_memory=True)


def test_from_mda():
    res16 = U.residues[16]
    ensemble = chilife.RotamerEnsemble.from_mda(res16, eval_clash=False)
    ensemble.save_pdb("test_data/test_from_MDA.pdb")

    with open("test_data/ans_from_MDA.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("test_data/test_from_MDA.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove("test_data/test_from_MDA.pdb")
    assert ans == test


def test_with_sample():
    np.random.seed(200)

    SL = chilife.SpinLabel("R1C", 28, ubq, sample=2000)

    with np.load('test_data/withsample.npz') as f:
        ans = {key: f[key] for key in f}
    
    np.testing.assert_almost_equal(SL._coords, ans['coords'])
    np.testing.assert_almost_equal(SL.weights, ans['weights'])
    np.testing.assert_almost_equal(SL.dihedrals, ans['dihedrals'])

    assert len(SL._coords) == len(SL.internal_coords)


def test_sample_partial():
    np.random.seed(200)
    SL = chilife.RotamerEnsemble("R1M", 28, ubq, sample=200, dihedral_sigmas=[0, 0, 0, np.inf, np.inf])


def test_update():
    SL1= chilife.RotamerEnsemble('R1M', 28, ubq, eval_clash=True)
    SL2 = chilife.RotamerEnsemble('R1M', 28, U, eval_clash=True)
    maxdiff = np.linalg.norm(SL1.origin - SL2.origin).max()
    assert maxdiff > 1

    SL1.protein = U
    SL1.update()

    np.testing.assert_allclose(SL1.coords, SL2.coords, rtol=1e-5)
    np.testing.assert_allclose(SL1.dihedrals, SL2.dihedrals, rtol=1e-5)
    np.testing.assert_allclose(SL1.weights, SL2.weights, rtol=1e-5)


def test_user_label():
    SL = chilife.SpinLabel("TRT", 28, ubq, "A")
    chilife.save("28TRT.pdb", SL, KDE=False)

    with open("test_data/28TRT.pdb", "r") as f:
        nohstring1 = "".join((line for line in f.readlines() if line[76:79].strip() != 'H'))
        ans = hashlib.md5(nohstring1.encode('utf-8')).hexdigest()

    with open("28TRT.pdb", "r") as f:
        nohstring2 = "".join((line for line in f.readlines() if line[76:79].strip() != 'H'))
        test = hashlib.md5(nohstring2.encode('utf-8')).hexdigest()

    os.remove("28TRT.pdb")

    assert test == ans


def test_save_pkl():
    Lys = chilife.RotamerEnsemble("LYS")

    with open("test_data/Lys.pkl", "wb") as f:
        pickle.dump(Lys, f)

    with open("test_data/Lys.pkl", "rb") as f:
        reload = pickle.load(f)

    os.remove("test_data/Lys.pkl")

    assert Lys == reload


def test_save_pkl_2():
    res16 = U.residues[16]
    ensemble = chilife.RotamerEnsemble.from_mda(res16)
    with open("test_data/res16.pkl", "wb") as f:
        pickle.dump(ensemble, f)

    with open("test_data/res16.pkl", "rb") as f:
        reload = pickle.load(f)

    assert ensemble == reload
    os.remove("test_data/res16.pkl")


def test_sample():
    np.random.seed(200)
    K48 = chilife.RotamerEnsemble.from_mda(ubq.residues[47])
    coords, weight = K48.sample(off_rotamer=True)

    wans = 0.0037019925960148086
    cans = np.array([[20.76406623, 27.90259642, 22.59445098],
                     [21.54999924, 26.79599953, 23.13299942],
                     [23.02842999, 27.03777002, 22.93937713],
                     [23.39710135, 27.56595972, 21.90105423],
                     [21.1101072, 25.47090301, 22.45317533],
                     [21.19895831, 25.51252905, 20.90717057],
                     [21.03430238, 24.15607743, 20.2156025],
                     [22.11572213, 23.19412351, 20.72492909],
                     [22.16827211, 22.00913727, 19.85099131]])

    np.testing.assert_almost_equal(coords, cans, decimal=6)
    assert weight == wans


def test_multisample():
    R1M = chilife.SpinLabel.from_wizard("R1M", site=48, protein=ubq, to_find=10000)


@pytest.mark.parametrize("res", chilife.SUPPORTED_RESIDUES)
def test_lib_distribution_persists(res):
    if res in chilife.USER_dLIBRARIES:
        return None
    elif res in chilife.USER_LIBRARIES:
        L1 = chilife.SpinLabel(res)
        L2 = chilife.SpinLabel(res, sample=100)
    else:
        L1 = chilife.RotamerEnsemble(res)
        L2 = chilife.RotamerEnsemble(res, sample=100)

    np.testing.assert_almost_equal(L1._rdihedrals, L2._rdihedrals)
    np.testing.assert_almost_equal(L1._rsigmas, L2._rsigmas)
    np.testing.assert_almost_equal(L1._weights, L2._weights)


methods = {"rosetta": np.array([[37.48119755, 25.73845177, 14.58725025],
                                [38.79399872, 25.76099968, 13.88000011],
                                [38.72872968, 26.58183116, 12.62501783],
                                [39.70399857, 27.34600067, 12.27700043]]),
           "bisect": np.array([[37.47976073, 25.7405822 , 14.58464153],
                               [38.79399872, 25.76099968, 13.88000011],
                               [38.73227959, 26.58109478, 12.62435692],
                               [39.70399857, 27.34600067, 12.27700043]]),
           "mmm": np.array([[37.47833189, 25.74271275, 14.5820287 ],
                            [38.79399872, 25.76099968, 13.88000011],
                            [38.73583004, 26.5803534 , 12.62370364],
                            [39.70399857, 27.34600067, 12.27700043]]),
           "fit": None}


@pytest.mark.parametrize(("method"), methods)
def test_alignment_method(method):
    if method == "fit":
        with pytest.raises(NotImplementedError) as e_info:
            SL = chilife.SpinLabel("R1C", site=28, protein=ubq, alignment_method=method, eval_clash=False)

    else:
        SL = chilife.SpinLabel("R1C", site=28, protein=ubq, alignment_method=method)
        np.testing.assert_allclose(SL.backbone, methods[method])




def test_catch_unused_kwargs():
    with pytest.raises(TypeError) as e_info:
        SL = chilife.SpinLabel("R1C", site=28, protein=ubq, fake_keyword="mmm")
    assert (
            str(e_info.value) == "Got unexpected keyword argument(s): fake_keyword"
    )


def test_guess_chain():
    anf = mda.Universe("test_data/1anf.pdb", in_memory=True)
    SL = chilife.SpinLabel.from_mmm("R1M", 20, energy_func=chilife.ljEnergyFunc(forgive=0.9))


@pytest.mark.parametrize(("resi", "ans"), ((20, "A"), (206, "B")))
def test_guess_chain2(resi, ans):
    SL = chilife.SpinLabel("R1C", resi, exou)
    assert SL.chain == ans


@pytest.mark.parametrize("resi", (100, 344))
def test_guess_chain_fail(resi):
    with pytest.raises(ValueError) as e_info:
        SL = chilife.SpinLabel("R1C", resi, exou)


def test_mem_sample():
    SL1 = chilife.SpinLabel('R1M', 28, ubq, sample=10000)
    print('booger')


def test_label_as_library():
    R1C = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, eval_clash=False)
    R1C_SL = chilife.SpinLabel("R1C", site=28, protein=ubq, eval_clash=False)
    np.testing.assert_equal(R1C.coords, R1C_SL.coords)
    np.testing.assert_equal(R1C.weights, R1C_SL.weights)
    np.testing.assert_equal(R1C.internal_coords.trajectory.coords, R1C_SL.internal_coords.trajectory.coords)


def test_coord_setter():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq)
    R1C2 = chilife.RotamerEnsemble("R1C", site=29, protein=ubq)

    R1C1.coords = R1C2.coords

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals, rtol=1e-6)

    for ic1, ic2 in zip(R1C1.internal_coords, R1C2.internal_coords):
        assert np.max(np.abs(np.cos(ic1.z_matrix[R1C1.ic_mask]) - np.cos(ic2.z_matrix[R1C2.ic_mask]))) < 0.03

    for ic1, ic2 in zip(R1C1.internal_coords, R1C2.internal_coords):
        assert np.max(np.abs(ic1.coords[R1C1.ic_mask] - ic2.coords[R1C2.ic_mask])) < 0.11

    for key in 'ori', 'mx':
        np.testing.assert_almost_equal(R1C1.internal_coords.chain_operators[0][key],
                                       R1C2.internal_coords.chain_operators[0][key])


def test_coord_setter2():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C2 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C1.coords = R1C2.coords[:, R1C2.side_chain_idx]

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals)


def test_dihedral_setter():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C2 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)

    R1C1.dihedrals = R1C2.dihedrals

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals)


def test_H_dihedral_setter():
    R1M = chilife.SpinLabel('R1M', 28, ubq, use_H=True)
    R1M2 = chilife.SpinLabel('R1M', 28, ubq, use_H=True)

    R1M.dihedrals = R1M.dihedrals + [180, 0, 0, 0, 0]

    Hidx = np.argwhere(R1M.atom_names == 'HA').flatten()

    np.testing.assert_allclose(R1M.coords[:, Hidx], R1M2.coords[:, Hidx])


def test_dihedral_setter_no_protein():
    R1M = chilife.SpinLabel('R1M', use_H=True)
    R1M2 = chilife.SpinLabel('R1M', use_H=True)
    R1M2.name += '2'

    R1M.dihedrals = R1M.dihedrals + [180, 0, 0, 0, 0]
    sorted_idxs = np.argsort(R1M.weights)[::-1]

    # Assert that the backbone does not move when
    np.testing.assert_allclose(R1M.aln, R1M2.aln)
    ic = R1M.internal_coords
    ic_angles = np.rad2deg([ic.get_dihedral(1, ['N', 'CA', 'CB', 'SG']) for ts in ic.trajectory])
    assert np.max(np.abs((ic_angles - R1M.dihedrals[:, 0])) % 360) < 1e-6
    assert np.max(np.abs(np.abs(ic_angles - R1M2.dihedrals[:, 0]) % 360  - 180)) < 1e-6




def test_get_sasa():
    R1C = chilife.RotamerEnsemble("R1C")
    sasas = R1C.get_sasa()
    sasans = np.load('test_data/sasas.npy')
    np.testing.assert_allclose(sasas, sasans)


def test_default_dihedral_sigmas():
    rotlib = chilife.read_bbdep('R1C', -60, -50)
    SL = chilife.SpinLabel('R1C')
    np.testing.assert_allclose(rotlib['sigmas'][rotlib['sigmas'] != 0], SL.sigmas[SL.sigmas != 35.])


def test_construct_with_dihedral_sigmas():
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=25)
    assert np.all(SL.sigmas == 25.)
    assert SL.sigmas.shape == (len(SL), len(SL.dihedral_atoms))


def test_construct_with_array_of_dihedral_sigmas():
    set_sigmas = [10, 20, 30, 40, 50]
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=set_sigmas)
    for i in range(5):
        assert np.all(SL.sigmas[:, i] == set_sigmas[i])

    assert SL.sigmas.shape == (len(SL), len(SL.dihedral_atoms))


def test_construct_with_full_array_of_dihedral_sigmas():
    set_sigmas = np.random.rand(148, 5)
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=set_sigmas)
    np.testing.assert_allclose(set_sigmas, SL.sigmas)


def test_dihedral_sigmas_fail():
    with pytest.raises(ValueError):
        SL = chilife.SpinLabel('R1C', dihedral_sigmas=[5, 2, 1])


def test_set_dihedral_sigmas():
    SL = chilife.SpinLabel('R1M')
    assert np.all(SL.sigmas == 35.)
    SL.set_dihedral_sampling_sigmas(25)
    assert np.all(SL.sigmas == 25.)


def test_clash_radius():
    RL = chilife.SpinLabel('TRT')
    assert RL.clash_radius == np.linalg.norm(RL.coords - RL.clash_ori, axis=-1).max() + 5


def test_nataa():
    SL = chilife.SpinLabel('R1M', 28, ubq)
    assert SL.nataa == 'A'


def test_from_trajectory():
    RE1 = chilife.RotamerEnsemble.from_trajectory(traj, 232, burn_in=0)
    RE2 = chilife.RotamerEnsemble.from_trajectory(traj, 231, burn_in=0)
    RE3 = chilife.RotamerEnsemble.from_trajectory(traj, 230, burn_in=0)

    assert len(RE1) == 1
    assert len(RE2) == 1
    assert len(RE3) == 10

    assert RE1.res == 'TRP'
    assert RE2.res == 'ALA'
    assert RE3.res == 'TRP'

    np.testing.assert_almost_equal(RE1.dihedrals, np.array([[-72.49934635, 164.56177856]]), decimal=5)
    np.testing.assert_almost_equal(RE2.dihedrals, np.array([[]]))
    np.testing.assert_almost_equal(RE3.dihedrals, np.array([[-176.38805 ,  -20.152264],
                                                            [-176.38805 ,  -52.83812 ],
                                                            [-176.38805 , -111.394165],
                                                            [-176.38805 ,   73.81698 ],
                                                            [-176.38805 , -134.54898 ],
                                                            [-176.38805 ,  118.280266],
                                                            [-176.38805 ,  164.5945  ],
                                                            [  62.16344 ,  -93.79344 ],
                                                            [ -69.9084  ,  -39.25944 ],
                                                            [ -69.9084  ,  161.67407 ]]), decimal=4)

    SL1 = chilife.SpinLabel.from_trajectory(traj, 238, burn_in=0, spin_atoms=['N1', 'O1'])
    assert np.all(SL1.spin_atoms == np.array(['N1', 'O1']))


def test_spin_from_traj():
    SL1 = chilife.RotamerEnsemble.from_trajectory(traj, 238, burn_in=0)
    np.testing.assert_equal(SL1.spin_atoms, ['N1', 'O1'])
    np.testing.assert_equal(SL1.spin_weights, [0.5, 0.5])


def test_from_traj_to_rotlib():
    U = chilife.load_protein('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    SL1 = chilife.RotamerEnsemble.from_trajectory(U, 2, chain='A')
    SL1.to_rotlib('___')
    SL2 = chilife.RotamerEnsemble('CYR', rotlib='___')
    os.remove('____rotlib.npz')

    # Ensure there is a rotamer for each frame
    assert len(SL2) == len(U.trajectory)

    # Ensure the rotamer backbones are aligned to the CA atom
    np.testing.assert_allclose(SL2.coords[:, 1], 0)

    # Ensure the dihedrals have not changed
    np.testing.assert_allclose(SL1.dihedrals, SL2.dihedrals)


def test_from_traj_dihedrals():
    SL1 = chilife.RotamerEnsemble.from_trajectory(traj, 238, burn_in=0)
    np.testing.assert_equal(SL1.dihedral_atoms, [['N', 'CA', 'CB', 'SG'],
                                                 ['CA', 'CB', 'SG', 'SD'],
                                                 ['CB', 'SG', 'SD', 'CE'],
                                                 ['SG', 'SD', 'CE', 'C3'],
                                                 ['SD', 'CE', 'C3', 'C4']])

def test_from_traj_user_dihedrals():
    U = chilife.load_protein('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    dihedral_atoms = [['N', 'CA', 'CB', 'SG'],
                      ['CA', 'CB', 'SG', 'S1L'],
                      ['CB', 'SG', 'S1L', 'C1L'],
                      ['SG', 'S1L', 'C1L', 'C1R'],
                      ['S1L', 'C1L', 'C1R', 'C1']]
    RL1 = chilife.RotamerEnsemble.from_trajectory(U, 2, chain='A', dihedral_atoms=dihedral_atoms)
    np.testing.assert_equal(RL1.dihedral_atoms, dihedral_atoms)

def test_from_traj_guess_dihedrals():
    U = chilife.load_protein('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    dihedral_atoms = [['N', 'CA', 'CB', 'SG'],
                      ['CA', 'CB', 'SG', 'S1L'],
                      ['CB', 'SG', 'S1L', 'C1L'],
                      ['SG', 'S1L', 'C1L', 'C1R'],
                      ['S1L', 'C1L', 'C1R', 'C2R']]
    RL1 = chilife.RotamerEnsemble.from_trajectory(U, 2, chain='A')
    np.testing.assert_equal(RL1.dihedral_atoms, dihedral_atoms)


def test_from_traj_mobile_bb():
    U = chilife.load_protein('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    RL1 = chilife.RotamerEnsemble.from_trajectory(U, 2, chain='A')
    test = np.squeeze(RL1.coords[:, RL1.aln_idx])
    ans = np.load('test_data/from_traj_mobile_bb.npy')
    np.testing.assert_almost_equal(test, ans)


def test_intra_fit():
    U = chilife.load_protein('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    RL1 = chilife.RotamerEnsemble.from_trajectory(U, 2, chain='A')
    RL1.intra_fit()
    bbs = np.squeeze(RL1.coords[:, RL1.aln_idx])
    bbref = RL1.aln

    assert np.all(np.linalg.norm(bbs - bbref, axis=(1, 2)) < 0.2)


def test_to_rotlib():

    # Generate rotamer library from trajectory
    RE = chilife.RotamerEnsemble.from_trajectory(traj, 230, burn_in=0)
    RE.to_rotlib('Test')
    
    # Build rotamer ensemble from new rotamer library
    RE2 = chilife.RotamerEnsemble('TRP', rotlib='Test')

    # Load new rotamer libary, then delete file
    with np.load(f"Test_rotlib.npz", allow_pickle=True) as f:
        rotlib_test = dict(f)

    os.remove('Test_rotlib.npz')

    # Load previously saved reference rotamer libary
    with np.load(f"test_data/Test_rotlib.npz", allow_pickle=True) as f:
        rotlib_reference = dict(f)

    assert RE2.res == 'TRP'
    assert len(RE2) == 10

    np.testing.assert_almost_equal(RE2.dihedrals, np.array([[-176.38805 ,  -20.152264],
                                                            [-176.38805 ,  -52.83812 ],
                                                            [-176.38805 , -111.394165],
                                                            [-176.38805 ,   73.81698 ],
                                                            [-176.38805 , -134.54898 ],
                                                            [-176.38805 ,  118.280266],
                                                            [-176.38805 ,  164.5945  ],
                                                            [  62.16344 ,  -93.79344 ],
                                                            [ -69.9084  ,  -39.25944 ],
                                                            [ -69.9084  ,  161.67407 ]]), decimal=4)

    np.testing.assert_almost_equal(rotlib_test['coords'], rotlib_reference['coords'], decimal=5)
    np.testing.assert_almost_equal(rotlib_test['weights'], rotlib_reference['weights'], decimal=5)
    np.testing.assert_almost_equal(rotlib_test['dihedrals'], rotlib_reference['dihedrals'], decimal=4)


def test_sample_persists():
    rot1 = chilife.RotamerEnsemble('ARG', 28, ubq)
    rot2 = chilife.RotamerEnsemble('ARG', 28, ubq, eval_clash=True)

    np.random.seed(10)
    x = rot1.sample()

    np.random.seed(10)
    y = rot2.sample()

    np.testing.assert_allclose(x[0], y[0])
    assert x[1] == y[1]


def test_trim_false():
    rot1 = chilife.RotamerEnsemble('ARG', 28, ubq)
    rot2 = chilife.RotamerEnsemble('ARG', 28, ubq, trim=False)
    rot3 = chilife.RotamerEnsemble('ARG', 28, ubq, eval_clash=False)

    assert len(rot2) == len(rot3)
    most_probable = np.sort(rot2.weights)[::-1][:len(rot1)]
    most_probable /= most_probable.sum()
    np.testing.assert_almost_equal(rot1.weights, most_probable)
    assert np.any(np.not_equal(rot2.weights, rot3.weights))

def test_min_callback():
    vals = []
    ivals = []
    def my_callback(val, i):
        vals.append(val)
        ivals.append(i)

    SL1 = chilife.RotamerEnsemble('ARG', 28, ubq)
    SL1.minimize(callback=my_callback)

    assert len(vals) > 0
    assert len(ivals) > 0


def test_copy_custom_lib():
    XYZ41 = chilife.SpinLabel('XYZ', 28, ubq, rotlib='test_data/usr_rtlb/XYZ_rotlib.npz')
    x2 = XYZ41.copy()

    assert x2 is not XYZ41
    assert x2.coords is not XYZ41.coords
    assert x2.internal_coords is not XYZ41.internal_coords
    assert x2.weights is not XYZ41.weights


def test_from_wizard_custom_rotlib():
    np.random.seed(0)
    A28NBA = chilife.SpinLabel.from_wizard('TES', 28, ubq, rotlib='test_data/usr_rtlb/NBA_rotlib.npz')
    ans = np.load('test_data/from_wiz_cust.npy')
    np.testing.assert_almost_equal(A28NBA.coords, ans, decimal=5)


def test_name():
    ubqs = ubq.select_atoms('resnum 8-70')
    SL1 = chilife.SpinLabel('R1M', 28, ubqs)
    assert SL1.name == 'A28R1M'


def test_ignore_waters():
    ub = chilife.fetch('1ubq')
    RL1 = chilife.RotamerEnsemble('ARG', 28, ub)
    RL2 = chilife.RotamerEnsemble('ARG', 28, ub, ignore_waters=False)

    assert len(RL1) != len(RL2)

    ans = np.load('test_data/ignore_waters.npy')
    np.testing.assert_almost_equal(RL2.coords, ans, decimal=5)


@pytest.mark.parametrize(('resn', 'ans'), (('HID', 17), ('HIP', 18), ('HIE', 17), ('ASH', 13),
                                           ('GLH', 16), ('CYM', 10), ('LYN', 21)))
def test_alt_prot(resn, ans):
    RL = chilife.RotamerEnsemble(resn, 28, ubq, use_H=True)
    assert len(RL.atom_names) == ans
    assert RL.res == resn

@pytest.mark.parametrize('res', ('HID', 'HIE', 'ASH', 'GLH', 'CYM', 'LYN', 'HIP'))
def test_from_mda_alt_prot(res):
    mda_residue = mda.Universe(f'test_data/{res.lower()}.pdb').residues[0]
    assert mda_residue.resname != res
    RL = chilife.RotamerEnsemble.from_mda(mda_residue, use_H=True)
    assert RL.res == res



import os, hashlib, pickle
import numpy as np
import pytest
import chilife
from chilife.MolSys import MolSys, parse_paren, Residue, ResidueSelection
import MDAnalysis as mda

prot = MolSys.from_pdb('test_data/1omp_H.pdb')
mda_prot = mda.Universe('test_data/1omp_H.pdb')


def test_from_pdb():
    ans = np.array([[-1.924, -20.646, -23.898],
                    [-1.825, -19.383, -24.623],
                    [-1.514, -19.5, -26.1],
                    [-2.237, -20.211, -26.799],
                    [-0.696, -18.586, -23.89]])

    np.testing.assert_almost_equal(prot.coords[21:26], ans)


def test_from_multistate_pdb():
    prot = MolSys.from_pdb('test_data/DHC.pdb')


def test_parse_paren1():
    P = parse_paren('(name CA or name CB)')
    assert P == ['name CA or name CB']


def test_parse_paren2():
    P = parse_paren('(resnum 32 and name CA) or (resnum 33 and name CB)')
    assert len(P) == 3


def test_parse_paren3():
    P = parse_paren('resnum 32 and name CA or (resnum 33 and name CB)')
    assert len(P) == 2


def test_parse_paren4():
    P = parse_paren('resname LYS ARG PRO and (name CA or (name CB and resname PRO)) or resnum 5')
    assert len(P) == 3


def test_parse_paren5():
    P = parse_paren('(name CA or (name CB and resname PRO) or name CD)')
    assert len(P) == 3


def test_select_or():
    m1 = prot.select_atoms('name CA or name CB')
    ans_mask = (prot.names == 'CA') + (prot.names == 'CB')

    np.testing.assert_allclose(m1.mask, np.argwhere(ans_mask).T[0])
    assert np.all(np.isin(m1.names, ['CA', 'CB']))
    mm = np.ones(prot.n_atoms, dtype=bool)
    mm[m1.mask] = False
    assert not np.any(np.isin(prot.names[mm], ['CA', 'CB']))


def test_select_and_or_and():
    m1 = prot.select_atoms('(resnum 32 and name CA) or (resnum 33 and name CB)')
    ans_mask = (prot.resnums == 32) * (prot.names == 'CA') + (prot.resnums == 33) * (prot.names == 'CB')

    np.testing.assert_almost_equal(m1.mask, np.argwhere(ans_mask).T[0])
    assert np.all(np.isin(m1.names, ['CA', 'CB']))
    assert np.all(np.isin(m1.resnums, [32, 33]))


def test_select_and_not():
    m1 = prot.select_atoms('resnum 32 and not name CA')
    ans_mask = (prot.resnums == 32) * (prot.names != 'CA')

    np.testing.assert_almost_equal(m1.mask, np.argwhere(ans_mask).T[0])
    assert not np.any(np.isin(m1.names, ['CA']))
    assert np.all(np.isin(m1.resnums, [32]))


def test_select_name_and_resname():
    m1 = prot.select_atoms("name CB and resname PRO")
    ans_mask = (prot.names == 'CB') * (prot.resnames == 'PRO')

    np.testing.assert_almost_equal(m1.mask, np.argwhere(ans_mask).T[0])
    assert np.all(np.isin(m1.names, ['CB']))
    assert np.all(np.isin(m1.resnames, ['PRO']))


def test_select_complex():
    m1 = prot.select_atoms('resname LYS ARG PRO and (name CA or (type C and resname PRO)) or resnum 5')

    resnames = np.isin(prot.resnames, ['LYS', 'ARG', 'PRO'])
    ca_or_c_pro = ((prot.names == 'CA') + ((prot.atypes == 'C') * (prot.resnames == 'PRO')))
    resnum5 = (prot.resnums == 5)
    ans_mask = resnames * ca_or_c_pro + resnum5

    np.testing.assert_almost_equal(m1.mask, np.argwhere(ans_mask).T[0])
    assert not np.any(np.isin(prot.resnums[~m1.mask], 5))


def test_select_range():
    m1 = prot.select_atoms('resnum 5-15')
    m2 = prot.select_atoms('resnum 5:15')
    ans_mask = np.isin(prot.resnums, list(range(5, 16)))
    np.testing.assert_almost_equal(m1.mask, np.argwhere(ans_mask).T[0])
    np.testing.assert_almost_equal(m2.mask, np.argwhere(ans_mask).T[0])


features = ("atomids", "names", "altlocs", "resnames", "chains", "resnums", "occupancies", "bs", "atypes", "charges")


@pytest.mark.parametrize('feature', features)
def test_AtomSelection_features(feature):
    m1 = prot.select_atoms('resname LYS ARG PRO and (name CA or (type C and resname PRO)) or resnum 5')
    A = prot.__getattr__(feature)[m1.mask]
    B = m1.__getattr__(feature)

    assert np.all(A == B)


def test_byres():
    waters = prot.select_atoms('byres name OH2 or resname HOH')
    assert np.all(np.unique(waters.resixs) == np.arange(371, 444, dtype=int))


def test_unary_not():
    asel = prot.select_atoms('not resid 5')
    ans = prot.resnums != 5
    assert np.all(asel.mask == np.argwhere(ans).T[0])


def test_around():
    asel = prot.select_atoms('around 5 resnum 5')
    mdasel = mda_prot.select_atoms('around 5 resnum 5')

    assert np.all(np.sort(asel.atoms.names) == np.sort(mdasel.atoms.names))
    assert np.all(asel.atoms.residues.resnums == mdasel.atoms.residues.resnums)


def test_subsel():
    asel = prot.select_atoms('around 5 resnum 5')
    bsel = prot.select_atoms('resname LEU around 5 resnum 5')
    a2sel = asel.select_atoms('resname LEU')
    np.testing.assert_almost_equal(bsel.coords, a2sel.coords)


def test_SL_Protein():
    omp = chilife.MolSys.from_pdb('test_data/1omp.pdb')
    ompmda = mda.Universe('test_data/1omp.pdb')

    SL1 = chilife.SpinLabel('R1M', 20, omp)
    SL2 = chilife.SpinLabel('R1M', 20, ompmda)

    assert SL1 == SL2


def test_trajectory():
    p = chilife.MolSys.from_pdb('test_data/2klf.pdb')
    ans = np.array([[41.068, 2.309, 0.296],
                    [39.431, 5.673, 2.669],
                    [39.041, 6.607, 2.738],
                    [39.385, 5.41, 3.503],
                    [35.887, 7.796, 3.6],
                    [34.155, 6.464, 8.545],
                    [39.874, 3.395, 2.951],
                    [35.945, 5.486, 1.808],
                    [39.028, 4.272, 1.541],
                    [40.395, 3.852, 9.391]])
    test = []
    for _ in p.trajectory:
        test.append(p.coords[0])
    test = np.array(test)

    np.testing.assert_allclose(test, ans)


def test_trajectory_set():
    p = chilife.MolSys.from_pdb('test_data/2klf.pdb')
    p.trajectory[5]
    np.testing.assert_allclose([34.155, 6.464, 8.545], p.coords[0])


def test_selection_traj():
    p = chilife.MolSys.from_pdb('test_data/2klf.pdb')
    s = p.select_atoms('resi 10 and name CB')
    np.testing.assert_allclose(s.coords, [[16.569, -1.406, 11.158]])
    p.trajectory[5]
    np.testing.assert_allclose(s.coords, [[16.195, -1.233, 11.306]])


def test_ResidueSelection():
    p = chilife.MolSys.from_pdb('test_data/1omp.pdb')
    r = p.residues[10:12]
    assert isinstance(r, ResidueSelection)
    assert len(r) == 2
    assert np.all(r.resnums == [11, 12])

    r2 = p.residues[10]
    assert isinstance(r2, Residue)
    assert r2.resname == 'ILE'
    assert r2.resnum == 11


def test_save_Protein():
    p = chilife.MolSys.from_pdb('test_data/1omp.pdb', sort_atoms=True)
    p._fname = "my_protein"
    chilife.save('my_protein.pdb', p)

    with open('test_data/test_save_xlprotein.pdb', 'r') as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("my_protein.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('my_protein.pdb')
    assert test == ans


def test_bool_index_atomsel():
    bindex = prot.resnames == 'LYS'
    x = prot.atoms[bindex]
    assert len(x) == 769
    assert np.all(x.resnames == 'LYS')


def test_save_trajectory():
    traj = chilife.MolSys.from_pdb('test_data/xlsavetraj.pdb')
    traj._fname = 'xlsavetraj'
    chilife.save('xlsavetraj.pdb', traj)

    with open('test_data/xlsavetraj.pdb', 'r') as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("xlsavetraj.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('xlsavetraj.pdb')
    assert test == ans


def test_trajectory_iter():
    traj = chilife.MolSys.from_pdb('test_data/test_traj_iter.pdb')
    assert not np.all(traj.trajectory.coords[0] == traj.trajectory.coords[1])

    prev_coords = np.zeros_like(traj.coords)
    for ts in traj.trajectory:
        this_coords = traj.coords
        assert not np.all(this_coords == prev_coords)
        prev_coords = this_coords.copy()


def test_xl_protein_repack():

    np.random.seed(2)
    protein = chilife.MolSys.from_pdb("test_data/1ubq.pdb").select_atoms("protein")
    SL = chilife.SpinLabel("R1M", site=28, protein=protein)

    traj1, deltaE1 = chilife.repack(protein, SL, repetitions=10, repack_radius=10)
    traj2, deltaE2 = chilife.repack(protein, SL, repetitions=10, off_rotamer=True, repack_radius=10)

    t1coords = traj1.universe.trajectory.coordinate_array
    t2coords = traj2.universe.trajectory.coordinate_array

    with np.load("test_data/repack_ans.npz") as f:
        t1ans = f["traj1"]
        t2ans = f["traj2"]

    np.testing.assert_almost_equal(t1coords, t1ans, decimal=4)
    np.testing.assert_almost_equal(t2coords, t2ans, decimal=4)


def test_same_as_mda():
    RE1 = chilife.RotamerEnsemble('ILE', 116, mda_prot)
    RE2 = chilife.RotamerEnsemble('ILE', 116, prot)

    np.testing.assert_almost_equal(RE1.coords, RE2.coords, decimal=5)
    np.testing.assert_almost_equal(RE1.weights, RE2.weights, decimal=5)

    SL1 = chilife.SpinLabel('R1M', 238, mda_prot)
    SL2 = chilife.SpinLabel('R1M', 238, prot)

    np.testing.assert_almost_equal(SL1.coords, SL2.coords, decimal=5)
    np.testing.assert_almost_equal(SL1.weights, SL2.weights, decimal=5)


def test_xl_protein_mutate():
    p = chilife.MolSys.from_pdb('test_data/1omp.pdb', sort_atoms=True)
    SL = chilife.SpinLabel('R1M', 238, p)
    pSL = chilife.mutate(p, SL)
    pSL._fname = 'test_mutate'
    chilife.save('mutate_xlprotein.pdb', pSL)

    with open('test_data/mutate_xlprotein.pdb', 'r') as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("mutate_xlprotein.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('mutate_xlprotein.pdb')
    assert ans == test


def test_re_form_xl_traj():
    traj = MolSys.from_pdb('test_data/xlsavetraj.pdb')
    SL = chilife.RotamerEnsemble.from_trajectory(traj, 236, burn_in=0)
    ans_dihedrals = np.array([[-175.57682647, -23.12323366],
                              [-175.57682647, 5.02763709],
                              [-175.57682647, 63.16005353],
                              [-175.57682647, 63.65448686],
                              [-175.57682647, 89.75074833]])

    assert len(SL) == 5
    np.testing.assert_almost_equal(SL.dihedrals, ans_dihedrals)


def test_backbone_selection():
    PhiSel = prot.residues[40].phi_selection()
    PsiSel = prot.residues[40].psi_selection()
    phi = np.rad2deg(chilife.get_dihedral(PhiSel.positions))
    psi = np.rad2deg(chilife.get_dihedral(PsiSel.positions))

    assert phi == -67.45821389595835
    assert psi == 135.37370203341246


# def test_xl_protein_from_pose():
#     assert False

def test_atom_sel_getitem():
    res = prot.residues[10].atoms[0:4]
    assert np.all(res.resnums == 11)
    np.testing.assert_equal(res.names, ['N', 'CA', 'C', 'O'])


def test_residue_sel_getitem():
    res = prot.residues[10:20]
    res10 = res.residues[0]

    assert res10.resid == 11
    assert np.all(res10.atoms.ix == np.arange(147, 166))
    assert np.all(res.residues[3:5].resnums == [14, 15])


def test_copy():
    prot2 = prot.copy()
    np.testing.assert_equal(prot2.coords, prot.coords)
    assert prot2.coords is not prot

def test_setattr1():
    prot2 = prot.copy()
    prot2.names = 'test'
    assert np.all(prot2.names == 'test')


def test_setattr2():
    prot2 = prot.copy()

    asel = prot2.select_atoms('resname ARG')
    mask = np.argwhere(prot2.resnames == 'ARG').flatten()

    asel.resnames = 'TES'
    assert np.all(asel.resnames == 'TES')
    assert np.all(prot2.resnames[mask] == 'TES')


def test_pickle():
    mol = MolSys.from_pdb('test_data/PPII_Capped.pdb')
    with open('tmp.pkl', 'wb') as f:
        pickle.dump(mol, f)

    del mol
    with open('tmp.pkl', 'rb') as f:
        mol = pickle.load(f)

    ans = MolSys.from_pdb('test_data/PPII_Capped.pdb')

    assert mol is not ans

    np.testing.assert_equal(mol.names, ans.names)
    np.testing.assert_equal(mol.resnames, ans.resnames)
    np.testing.assert_equal(mol.coords, ans.coords)


def test_pickle_selection():
    mol = MolSys.from_pdb('test_data/PPII_Capped.pdb').atoms
    with open('tmp.pkl', 'wb') as f:
        pickle.dump(mol, f)

    del mol
    with open('tmp.pkl', 'rb') as f:
        mol = pickle.load(f)

    ans = MolSys.from_pdb('test_data/PPII_Capped.pdb').atoms

    assert mol is not ans

    np.testing.assert_equal(mol.names, ans.names)
    np.testing.assert_equal(mol.resnames, ans.resnames)
    np.testing.assert_equal(mol.coords, ans.coords)


def test_from_atomsel():
    # MDAnalysis test
    atomsel = mda_prot.select_atoms('resnum 30-150')
    new_prot = MolSys.from_atomsel(atomsel)

    np.testing.assert_equal(new_prot.positions, atomsel.positions)
    np.testing.assert_equal(new_prot.ix + 435, atomsel.ix)
    assert len(new_prot.trajectory) == len(atomsel.universe.trajectory)

    # chiLife MolSys test
    atomsel = prot.select_atoms('resnum 30-150')
    new_prot = MolSys.from_atomsel(atomsel)

    np.testing.assert_equal(new_prot.positions, atomsel.positions)
    np.testing.assert_equal(new_prot.ix + 435, atomsel.ix)
    assert len(new_prot.trajectory) == len(atomsel.universe.trajectory)


def test_ResidueSelection_iter():
    prot = MolSys.from_pdb('test_data/1a2w.pdb')
    chain_B = prot.select_atoms('chain B')
    for res in chain_B.residues:
        assert res.chain == 'B'
        assert np.all(res.atoms.chains == 'B')
        

def test_ResidueSelection_iter2():
    protein = chilife.MolSys.from_pdb("test_data/1ubq.pdb").select_atoms("protein")
    ids = [res.resid for res in protein.residues]
    np.testing.assert_equal(ids, np.arange(1,77))






import numpy as np
from chilife.math_utils import *


def test_normalize_angle():
    angles = [-np.pi, 0, np.pi, 3 * np.pi/2, 2*np.pi]
    ans = [-np.pi, 0, np.pi, -np.pi/2, 0]
    norm_angles = normalize_angles(angles)
    np.testing.assert_almost_equal(norm_angles, ans)


def test_angle_dist():
    angles = np.linspace(0, 2*np.pi, 100)
    ref = np.zeros(100)
    ans = normalize_angles(angles)
    dist = angle_dist(angles, ref)
    np.testing.assert_almost_equal(dist, ans)

    # Rotate 90 degrees
    angles += np.pi/2
    ref += np.pi/2
    dist = angle_dist(angles, ref)
    np.testing.assert_almost_equal(dist, ans)



from typing import Tuple, Dict, Union, BinaryIO, TextIO, Protocol
import warnings
import os
import urllib
import MDAnalysis
from numpy.typing import ArrayLike
from collections import defaultdict
from hashlib import sha256
from pathlib import Path
import pickle
import shutil
from io import StringIO, BytesIO
import zipfile

import numpy as np
from scipy.stats import gaussian_kde
from memoization import cached, suppress_warnings
import MDAnalysis as mda

import chilife.RotamerEnsemble as re
import chilife.SpinLabel as sl
import chilife.dRotamerEnsemble as dre
from .globals import dihedral_defs, rotlib_indexes, RL_DIR, SUPPORTED_BB_LABELS, USER_RL_DIR, rotlib_defaults, alt_prot_states
from .alignment_methods import local_mx
from .IntrinsicLabel import IntrinsicLabel
from .MolSys import MolecularSystemBase
from .MolSysIC import MolSysIC

#                 ID    name   res  chain resnum      X     Y      Z      q      b              elem
fmt_str = "ATOM  {:5d} {:^4s} {:3s} {:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:>2s}  \n"


def read_distance_distribution(file_name: str) -> Tuple[np.ndarray, np.ndarray]:
    """Reads a DEER distance distribution file in the DeerAnalysis or similar format.

    Parameters
    ----------
    file_name : str
        File name of distance distribution

    Returns
    -------
    r: np.ndarray
        Domain of distance distribution in the same units as the file.
    p: np.ndarray
        Probability density over r normalized such that the integral of p over r is 1.
    """

    # Load DA file
    data = np.loadtxt(file_name)

    # Convert nm to angstroms
    r = data[:, 0]
    if max(r) < 15:
        r *= 10

    # Extract distance domain coordinates
    p = data[:, 1]
    return r, p


def hash_file(file: Union[Path, BinaryIO]):
    """
    Helper function for hashing rotamer library files

    Parameters
    ----------
    file : str, Path, BinaryIO
        The name, Path or IOBytes of the file to hash

    Returns
    -------
    hashstring : str
        file hash
    """
    hash = sha256()
    with open(file, 'rb') as f:
        while True:
            block = f.read(hash.block_size)
            if not block:
                break
            hash.update(block)

    return hash.hexdigest()


def get_possible_rotlibs(rotlib: str,
                         suffix: str,
                         extension: str,
                         return_all: bool = False,
                         was_none: bool = False) -> Union[Path, None]:
    """
    Search all known rotlib directories and the current working directory for rotlib(s) that match the provided
    information.

    Parameters
    ----------
    rotlib: str
        Fullname, base name or partial name of the rotamer libraries to search for.
    suffix: str
        possible suffixes the rotamer library may have, e.g. ip2 to indicate an i+2 rotamer library.
    extension: str
        filetype extension to look for. This will be either `npz` for monofunctional rotlibs or zip for bifunctional
        rotlibs.
    return_all: bool
        By default, only the first found rotlib will be returned unless ``return_all = True``, in which case
    was_none: bool
        For internal use only.

    Returns
    -------
    rotlib: Path, List[Path], None
        The path to the found rotlib or a list of paths to the rotlibs that match the search criteri, or ``None`` if
        no rotlibs are found.
    """
    cwd = Path.cwd()
    sufplusex = '_' + suffix + extension
    # Assemble a list of possible rotlib paths starting in the current directory
    possible_rotlibs = [Path(rotlib),
                        cwd / rotlib,
                        cwd / (rotlib + extension),
                        cwd / (rotlib + sufplusex)]

    possible_rotlibs += list(cwd.glob(f'{rotlib}*{sufplusex}'))
    # Then in the user defined rotamer library directory
    for pth in USER_RL_DIR:
        possible_rotlibs += list(pth.glob(f'{rotlib}*{sufplusex}'))

    if not was_none:
        possible_rotlibs += list((RL_DIR / 'user_rotlibs').glob(f'*{rotlib}*'))

    if return_all:
        rotlib = []
    for possible_file in possible_rotlibs:
        if possible_file.exists() and return_all:
                rotlib.append(possible_file)
        elif possible_file.exists() and not possible_file.is_dir():
            rotlib = possible_file
            break
    else:
        if isinstance(rotlib, str) and was_none and rotlib in rotlib_defaults:
            rotlib = RL_DIR / 'user_rotlibs' / (rotlib_defaults[rotlib][0] + sufplusex)

        elif not isinstance(rotlib, list) or rotlib == []:
            rotlib = None

    # rotlib lists need to be sorted to prevent position mismatches for results with tests.
    if isinstance(rotlib, list):
        rotlib = list(set(rotlib))
        rotlib = [rot for rot in rotlib if str(rot).endswith(extension)]
        rotlib = sorted(rotlib)
    else:
        rotlib = rotlib if str(rotlib).endswith(extension) else None

    return rotlib


suppress_warnings()
@cached(custom_key_maker=hash_file)
def read_rotlib(rotlib: Union[Path, BinaryIO] = None) -> Dict:
    """Reads RotamerEnsemble for stored spin labels.

    Parameters
    ----------
    rotlib : Path, BinaryIO
        Path object pointing to the rotamer library. Or a BytesIO object containing the rotamer library file
    Returns
    -------
    lib: dict
        Dictionary of SpinLabel rotamer ensemble attributes including coords, weights, dihedrals etc.

    """
    with np.load(rotlib, allow_pickle=True) as files:
        if files['format_version'] <= 1.1:
            raise RuntimeError('The rotlib that was provided is an old version that is not compatible with your '
                               'version of chiLife. You can either remake the rotlib, or use the update_rotlib.py '
                               'script provided in the chilife scripts directory to update this rotamer library to the '
                               'new format.')

        lib = dict(files)

    del lib["allow_pickle"]

    if "sigmas" not in lib:
        lib["sigmas"] = np.array([])
    lib['internal_coords'] = lib['internal_coords'].item()
    lib["_rdihedrals"] = np.deg2rad(lib["dihedrals"])
    lib["_rsigmas"] = np.deg2rad(lib["sigmas"])
    lib['rotlib'] = str(lib['rotlib'])
    lib['type'] = str(lib['type'])
    lib['format_version'] = float(lib['format_version'])
    return lib


@cached(custom_key_maker=hash_file)
def read_drotlib(rotlib: Path) -> Tuple[dict]:
    """Reads RotamerEnsemble for stored spin labels.

        Parameters
        ----------
        rotlib : Path
            The Path to the rotamer library file.
        Returns
        -------
        lib: Tuple[dict]
            Dictionaries of rotamer library attributes including sub_residues .
        """

    with zipfile.ZipFile(rotlib, 'r') as archive:
        for f in archive.namelist():
            if 'csts' in f:
                with archive.open(f) as of:
                    csts = np.load(of)
            elif f[-12] == 'A':
                with archive.open(f) as of:
                    libA = read_rotlib.__wrapped__(of)
            elif f[-12] == 'B':
                with archive.open(f) as of:
                    libB = read_rotlib.__wrapped__(of)

    return libA, libB, csts


@cached
def read_bbdep(res: str, Phi: int, Psi: int) -> Dict:
    """Read the Dunbrack rotamer library for the provided residue and backbone conformation.

    Parameters
    ----------
    res : str
        3-letter residue code
    Phi : int
        Backbone Phi dihedral angle for the provided residue
    Psi : int
        Backbone Psi dihedral angle for the provided residue

    Returns
    -------
    lib: dict
        Dictionary of arrays containing rotamer library information in cartesian and dihedral space
    """

    ic_res = alt_prot_states.get(res, res)

    lib = {}
    Phi, Psi = str(Phi), str(Psi)

    # Read residue internal coordinate structure
    with open(RL_DIR / f"residue_internal_coords/{res.lower()}_ic.pkl", "rb") as f:
        ICs = pickle.load(f)

    atom_types = ICs.atom_types.copy()
    atom_names = ICs.atom_names.copy()

    maxchi = 5 if res in SUPPORTED_BB_LABELS else 4
    nchi = np.minimum(len(dihedral_defs[res]), maxchi)

    if res not in ("ALA", "GLY"):
        library = "R1C.lib" if res in SUPPORTED_BB_LABELS else "ALL.bbdep.rotamers.lib"
        start, length = rotlib_indexes[f"{ic_res}  {Phi:>4}{Psi:>5}"]

        with open(RL_DIR / library, "rb") as f:
            f.seek(start)
            rotlib_string = f.read(length).decode()
            s = StringIO(rotlib_string)
            s.seek(0)
            data = np.genfromtxt(s, usecols=range(maxchi + 4, maxchi + 5 + 2 * maxchi))

        lib["weights"] = data[:, 0]
        lib["dihedrals"] = data[:, 1: nchi + 1]
        lib["sigmas"] = data[:, maxchi + 1: maxchi + nchi + 1]
        dihedral_atoms = dihedral_defs[res][:nchi]

        # Calculate cartesian coordinates for each rotamer
        z_matrix = ICs.batch_set_dihedrals(np.zeros(len(lib['dihedrals']), dtype=int), np.deg2rad(lib['dihedrals']), 1, dihedral_atoms)
        ICs._chain_operators = ICs._chain_operators[0]
        ICs.load_new(np.array(z_matrix))
        internal_coords = ICs.copy()
        coords = ICs.protein.trajectory.coordinate_array.copy()

    else:
        lib["weights"] = np.array([1])
        lib["dihedrals"], lib["sigmas"], dihedral_atoms = [], [], []
        coords = ICs.to_cartesian()[None, ...]
        internal_coords = ICs.copy()

    # Get origin and rotation matrix of local frame
    mask = np.in1d(atom_names, ["N", "CA", "C"])
    ori, mx = local_mx(*coords[0, mask])

    # Set coords in local frame and prepare output
    coords -= ori

    lib["coords"] = np.einsum('ijk,kl->ijl', coords, mx)
    lib["internal_coords"] = internal_coords
    lib["atom_types"] = np.asarray(atom_types, dtype=str)
    lib["atom_names"] = np.asarray(atom_names, dtype=str)
    lib["dihedral_atoms"] = np.asarray(dihedral_atoms, dtype=str)
    lib["_rdihedrals"] = np.deg2rad(lib["dihedrals"])
    lib["_rsigmas"] = np.deg2rad(lib["sigmas"])
    lib['rotlib'] = res
    lib['backbone_atoms'] = ["H", "N", "CA", "HA", "C", "O"]
    lib['aln_atoms'] = ['N', 'CA', 'C']

    # hacky solution to experimental backbone dependent rotlibs
    if res == 'R1C':
        lib['spin_atoms'] = np.array(['N1', 'O1'])
        lib['spin_weights'] = np.array([0.5, 0.5])

    return lib


def read_library(rotlib: str, Phi: float = None, Psi: float = None) -> Dict:
    """Function to read rotamer libraries and bifunctional rotamer libraries as dictionaries.

    Parameters
    ----------
    rotlib : str, Path
        3 letter residue code or path the rotamer library file.
    Phi : float
        Backbone Phi dihedral angle for the provided residue. Only applicable for canonical amino acids.
    Psi : float
        Backbone Psi dihedral angle for the provided residue. Only applicable for canonical amino acids.

    Returns
    -------
    lib: dict
        Dictionary of arrays containing rotamer library information in cartesian and dihedral space.
    """
    backbone_exists = Phi and Psi

    if backbone_exists:
        Phi = int((Phi // 10) * 10)
        Psi = int((Psi // 10) * 10)
    # Use helix backbone if not otherwise specified
    else:
        Phi, Psi = -60, -50

    if isinstance(rotlib, Path):
        if rotlib.suffix == '.npz':
            return read_rotlib(rotlib)
        elif rotlib.suffix == '.zip':
            return read_drotlib(rotlib)
        else:
            raise ValueError(f'{rotlib.name} is not a valid rotamer library file type')
    elif isinstance(rotlib, str):
        return read_bbdep(rotlib, Phi, Psi)
    else:
        raise ValueError(f'{rotlib} is not a valid rotamer library')


def save(
        file_name: str,
        *molecules: Union[re.RotamerEnsemble, MolecularSystemBase, mda.Universe, mda.AtomGroup, str],
        protein_path: Union[str, Path] = None,
        mode: str = 'w',
        conect: bool = False,
        frames: Union[int, str, ArrayLike, None] = 'all',
        **kwargs,
) -> None:
    """Save a pdb file of the provided labels and proteins

    Parameters
    ----------
    file_name : str
        Desired file name for output file. Will be automatically made based off of protein name and labels if not
        provided.
    *molecules : RotmerLibrary, chiLife.MolSys, mda.Universe, mda.AtomGroup, str
        Object(s) to save. Includes RotamerEnsemble, SpinLabels, dRotamerEnsembles, proteins, or path to protein pdb.
        Can add as many as desired, except for path, for which only one can be given.
    protein_path : str, Path
        The Path to a pdb file to use as the protein object.
    mode : str
        Which mode to open the file in. Accepts 'w' or 'a' to overwrite or append.
    conect : bool
        Whether to save  PDB CONECT records. Defaults to False.
    frames : int, str, ArrayLike[int], None
        Exclusively used for ``MolSys`` and ``AtomGroup`` objects. ``frames`` are the indices of the frames to save.
        Can be a single index, array of indices, 'all', or ``None``. If ``None`` is passed only the
        activate frame will be saved. If 'all' is passed, all frames are saved. Default is 'all'
    **kwargs :
        Additional arguments to pass to ``write_labels``

        write_spin_centers : bool
            Write spin centers (atoms named NEN) as a separate object with weights mapped to q-factor.

    """

    if isinstance(file_name, tuple(molecule_class.keys())):
        molecules = list(molecules)
        molecules.insert(0, file_name)
        file_name = None

    # Check for filename at the beginning of args
    tmolecules = defaultdict(list)
    for mol in molecules:
        mcls = [val for key, val in molecule_class.items() if isinstance(mol, key)]
        if mcls == []:
            raise TypeError(f'{type(mol)} is not a supported type for this function. '
                            'chiLife can only save objects of the following types:\n'
                            + ''.join(f'{key.__name__}\n' for key in molecule_class) +
                            'Please check that your input is compatible')

        tmolecules[mcls[0]].append(mol)
    molecules = tmolecules

    # Ensure only one protein path was provided (for now)
    protein_path = [] if protein_path is None else [protein_path]
    if len(protein_path) > 1:
        raise ValueError('More than one protein path was provided. C')
    elif len(protein_path) == 1:
        protein_path = protein_path[0]
    else:
        protein_path = None

    # Create a file name from protein information
    if file_name is None:
        if protein_path is not None:
            f = Path(protein_path)
            file_name = f.name[:-4]
        else:
            file_name = ""
            for protein in molecules['molcart']:
                if getattr(protein, "filename", None):
                    file_name += ' ' + Path(protein.filename).name
                    file_name = file_name[:-4]

        if file_name == '' and len(molecules['molcart']) > 0:
            file_name = "No_Name_Protein"

        # Add spin label information to file name
        if 0 < len(molecules['rotens']) < 3:
            naml = [file_name] + [rotens.name for rotens in molecules['rotens']]
            file_name = "_".join(naml)
            file_name = file_name.strip('_')
        elif len(molecules['rotens']) >= 3:
            file_name += "_many_labels"

        file_name += ".pdb"
        file_name = file_name.strip()

    if protein_path is not None:
        print(protein_path, file_name)
        shutil.copy(protein_path, file_name)
        pdb_file = open(file_name, 'a+')
    else:
        pdb_file = open(file_name, mode)

    used_names = {}
    for protein in molecules['molcart']:

        if isinstance(protein, (mda.AtomGroup, mda.Universe)):
            name = Path(protein.universe.filename) if protein.universe.filename is not None else Path(pdb_file.name)
            name = name.name
        else:
            name = protein.fname if hasattr(protein, 'fname') else None

        if name is None:
            name = Path(pdb_file.name).name

        name = name[:-4] if name.endswith(".pdb") else name
        name_ = name + str(used_names[name]) if name in used_names else name

        used_names[name] = used_names.setdefault(name, 0) + 1

        write_protein(pdb_file, protein, name_, conect=conect, frames=frames)

    for ic in molecules['molic']:
        if isinstance(ic.protein, (mda.AtomGroup, mda.Universe)):
            name = Path(ic.protein.universe.filename) if ic.protein.universe.filename is not None else Path(pdb_file.name)
            name = name.name
        else:
            name = ic.protein.fname if hasattr(ic.protein, 'fname') else None

        if name is None:
            name = Path(pdb_file.name).name

        name = name[:-4] if name.endswith(".pdb") else name
        name_ = name + str(used_names[name]) if name in used_names else name

        write_ic(pdb_file, ic, name=name_, conect=conect, frames=frames)

    if len(molecules['rotens']) > 0:
        write_labels(pdb_file, *molecules['rotens'], conect=conect, **kwargs)

    pdb_file.close()


def fetch(accession_number: str, save: bool = False) -> MDAnalysis.Universe:
    """Fetch pdb file from the protein data bank or the AlphaFold Database and optionally save to disk.

    Parameters
    ----------
    accession_number : str
        4 letter structure PDB ID or alpha fold accession number. Note that AlphaFold accession numbers must begin with
        'AF-'.
    save : bool
        If true the fetched PDB will be saved to the disk.

    Returns
    -------
    U : MDAnalysis.Universe
        MDAnalysis Universe object of the protein corresponding to the provided PDB ID or AlphaFold accession number

    """
    accession_number = accession_number.split('.pdb')[0]
    pdb_name = accession_number + '.pdb'

    if accession_number.startswith('AF-'):
        print(f"https://alphafold.ebi.ac.uk/files/{accession_number}-F1-model_v3.pdb")
        urllib.request.urlretrieve(f"https://alphafold.ebi.ac.uk/files/{accession_number}-F1-model_v3.pdb", pdb_name)
    else:
        urllib.request.urlretrieve(f"http://files.rcsb.org/download/{pdb_name}", pdb_name)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        U = mda.Universe(pdb_name, in_memory=True)

    if not save:
        os.remove(pdb_name)

    return U


def load_protein(struct_file: Union[str, Path],
                 *traj_file: Union[str, Path]) -> MDAnalysis.AtomGroup:
    """

    Parameters
    ----------
    struct_file : Union[TextIO, str, Path]
        Name, Path or TextIO object referencing the structure file (e.g. pdb, gro, psf)
    *traj_file : Union[TextIO, str, Path] (optional)
        Name, Path or TextIO object(s) referencing the trajectory file (e.g. pdb, xtc, dcd)


    Returns
    -------
    protein: MDAnalysis.AtomGroup
        An MDA AtomGroup object containing the protein structure and trajectory. The object is always loaded into
        memory to allow coordinate manipulations.
    """

    if traj_file != []:
        traj_file = [str(file) for file in traj_file]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            protein = mda.Universe(str(struct_file), *traj_file, in_memory=True)
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            protein = mda.Universe(struct_file, in_memory=True)

    return protein


def write_protein(pdb_file: TextIO,
                  protein: Union[mda.Universe, mda.AtomGroup, MolecularSystemBase],
                  name: str = None,
                  conect: bool = False,
                  frames: Union[int, str, ArrayLike, None] = 'all') -> None:
    """
    Helper function to write protein PDBs from MDAnalysis and MolSys objects.

    Parameters
    ----------
    pdb_file : TextIO
        File to save the protein to
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup, MolSys
        MDAnalysis or MolSys object to save
    name : str
        Name of the protein to put in the header
    frames : int, str, ArrayLike or None
        Frames of the trajectory/ensemble to save to file
    """

    # Change chain identifier if longer than 1
    available_segids = iter('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    for seg in protein.segments:
        if len(seg.segid) > 1:
            seg.segid = next(available_segids)

    traj = protein.universe.trajectory
    pdb_file.write(f'HEADER {name}\n')
    frames = None if frames == 'all' and len(traj) == 1 else frames

    if frames == 'all':
        for mdl, ts in enumerate(traj):
           write_frame(pdb_file, protein.atoms, mdl)
    elif hasattr(frames, '__len__'):
        for mdl, frame in enumerate(frames):
            traj[frame]
            write_frame(pdb_file, protein.atoms, mdl)
    elif frames is not None:
        traj[frames]
        write_frame(pdb_file, protein.atoms)
    else:
        write_frame(pdb_file, protein.atoms)

    if conect:
        bonds = (protein.bonds.indices if hasattr(protein, 'bonds') else
                protein.topology.bonds if hasattr(protein.topology, 'bonds') else
                None)

        if bonds is not None:
            write_bonds(pdb_file, bonds)


def write_frame(pdb_file: TextIO, atoms, frame=None, coords=None):

    if frame is not None:
        pdb_file.write(f"MODEL {frame + 1}\n")

    if coords is None:
        coords = atoms.positions

    [pdb_file.write(
        fmt_str.format(
            i + 1,
            atom.name,
            atom.resname[:3],
            atom.segid,
            atom.resnum,
            *coord,
            1.00,
            1.0,
            atom.type,
        )
    ) for i, (atom, coord) in enumerate(zip(atoms, coords))]

    pdb_file.write("TER\n")

    if frame is not None:
        pdb_file.write(f"ENDMDL\n")


def write_ic(pdb_file: TextIO,
             ic: MolSysIC,
             name: str = None,
             conect: bool = None,
             frames: Union[int, str, ArrayLike, None] = 'all')-> None:
    """
    Write a :class:`~MolSysIC` internal coordinates object to a pdb file.

    Parameters
    ----------
    pdb_file : TextIO
        open file or io object.
    ic: MolSysIC
        chiLife internal coordinates object.
    name : str
        Name of the molecule to be used for the HEADER.
    conect : bool, optional
        Write PDB CONECT information to file.
    frames : int, str, ArrayLike or None
        Frames of the trajectory/ensemble to save to file
    """
    frames = None if frames == 'all' and len(ic.trajectory) == 1 else frames
    pdb_file.write(f'HEADER {name}\n')
    traj = ic.trajectory
    if frames == 'all':
        for mdl, ts in enumerate(traj):
            write_frame(pdb_file, ic.atoms, mdl, ic.coords)
    elif hasattr(frames, '__len__'):
        for mdl, frame in enumerate(frames):
            traj[frame]
            write_frame(pdb_file, ic.atoms, mdl, ic.coords)
    elif frames is not None:
        traj[frames]
        write_frame(pdb_file, ic.atoms, coords=ic.coords)
    else:
        write_frame(pdb_file, ic.atoms, coords=ic.coords)

    if conect:
        bonds = ic.topology.bonds
        write_bonds(pdb_file, bonds)


def write_labels(pdb_file: TextIO, *args: sl.SpinLabel,
                 KDE: bool = True,
                 sorted: bool = True,
                 write_spin_centers: bool = True,
                 conect: bool = False) -> None:
    """Lower level helper function for saving SpinLabels and RotamerEnsembles. Loops over SpinLabel objects and appends
    atoms and electron coordinates to the provided file.

    Parameters
    ----------
    pdb_file : str
        File name to write to.
    *args: SpinLabel, RotamerEnsemble
        The RotamerEnsemble or SpinLabel objects to be saved.
    KDE: bool
        Perform kernel density estimate smoothing on rotamer weights before saving. Usefull for uniformly weighted
        RotamerEnsembles or RotamerEnsembles with lots of rotamers
    sorted : bool
        Sort rotamers by weight before saving.
    write_spin_centers : bool
        Write spin centers (atoms named NEN) as a seperate object with weights mapped to q-factor.

    Returns
    -------
    None
    """

    # Write spin label models
    for k, label in enumerate(args):
        pdb_file.write(f"HEADER {label.name}\n")

        # Save models in order of weight

        sorted_index = np.argsort(label.weights)[::-1] if sorted else np.arange(len(label.weights))
        norm_weights = label.weights / label.weights.max()
        if isinstance(label, dre.dRotamerEnsemble):
            sites = np.concatenate([np.ones(len(label.RE1.atoms), dtype=int) * int(label.site1),
                                    np.ones(len(label.RE2.atoms), dtype=int) * int(label.site2)])
        else:
            sites = [atom.resi for atom in label.atoms]
        for mdl, (conformer, weight) in enumerate(
                zip(label.coords[sorted_index], norm_weights[sorted_index])
        ):
            pdb_file.write(f"MODEL {mdl}\n")

            [
                pdb_file.write(
                    fmt_str.format(
                        i,
                        label.atom_names[i],
                        label.res[:3],
                        label.chain,
                        sites[i],
                        *conformer[i],
                        weight,
                        1.00,
                        label.atom_types[i],
                    )
                )
                for i in range(len(label.atom_names))
            ]
            pdb_file.write("TER\n")
            pdb_file.write("ENDMDL\n")
        if conect:
            write_bonds(pdb_file, label.bonds)

    # Write electron density at electron coordinates
    for k, label in enumerate(args):
        if not hasattr(label, "spin_centers"):
            continue
        if not np.any(label.spin_centers):
            continue

        pdb_file.write(f"HEADER {label.name}_density\n")
        spin_centers = np.atleast_2d(label.spin_centers)

        if KDE and len(spin_centers) > 5:
            try:
                # Perform gaussian KDE to determine electron density
                gkde = gaussian_kde(spin_centers.T, weights=label.weights)

                # Map KDE density to pseudoatoms
                vals = gkde.pdf(spin_centers.T)

            except:
                vals = label.weights

        else:
            vals = label.weights

        if write_spin_centers:
            norm_weights = vals / vals.max()
            [
                pdb_file.write(
                    fmt_str.format(
                        i,
                        "NEN",
                        label.label[:3],
                        label.chain,
                        int(label.site),
                        *spin_centers[i],
                        norm_weights[i],
                        1.00,
                        "N",
                    )
                )
                for i in range(len(norm_weights))
            ]

        pdb_file.write("TER\n")


def write_atoms(file: TextIO,
                atoms: Union[MDAnalysis.Universe, MDAnalysis.AtomGroup, MolecularSystemBase],
                coords: ArrayLike = None) -> None:
    """
    Write a set of atoms to a file in PDB format without any prefix or suffix, i.e. Not wrapped in `MODEL` or `ENDMDL`

    Parameters
    ----------
    file : TextIO
        A writable file (TextIO) object. The file that the atoms will be written to.
    atoms : MDAnalysis.Universe, MDAnalysis.AtomGroup, MolecularSystemBase
        Object containing atoms to be written. May also work with duck-typed set of atoms that are iterable and each
        atom has properties ``index``, ``name``,
    coords : ArrayLike
        Array of atom coordinates corresponding to atoms
    """

    if isinstance(atoms, MDAnalysis.Universe):
        atoms = atoms.atoms

    if coords is None:
        coords = atoms.positions

    for atom, coord in zip(atoms, coords):
        file.write(
            f"ATOM  {atom.index + 1:5d} {atom.name:^4s} {atom.resname:3s} {'A':1s}{atom.resnum:4d}    "
            f"{coord[0]:8.3f}{coord[1]:8.3f}{coord[2]:8.3f}{1.0:6.2f}{1.0:6.2f}          {atom.type:>2s}  \n"
        )


def write_bonds(pdb_file, bonds):
    active_atom = bonds[0][0]
    line = "CONECT" + f"{active_atom:>5d}"

    lines = []
    for b in bonds:
        if b[0] != active_atom or len(line) > 30:
            lines.append(line + '\n')
            active_atom = b[0]
            line = "CONECT" + f"{active_atom:>5d}"

        line += f"{b[1]:>5d}"
    lines.append(line + '\n')
    pdb_file.writelines(lines)


molecule_class = {re.RotamerEnsemble: 'rotens',
                  dre.dRotamerEnsemble: 'rotens',
                  IntrinsicLabel: 'rotens',
                  mda.Universe: 'molcart',
                  mda.AtomGroup: 'molcart',
                  MolecularSystemBase: 'molcart',
                  MolSysIC: 'molic'}


rotlib_formats = {1.0: (
    'rotlib',  #
    'resname',
    'coords',
    'internal_coords',
    'weights',
    'atom_types',
    'atom_names',
    'dihedrals',
    'dihedral_atoms',
    'type',
    'format_version'
)}

rotlib_formats[1.1] = *rotlib_formats[1.0], 'description', 'comment', 'reference'
rotlib_formats[1.2] = rotlib_formats[1.0]
rotlib_formats[1.3] = rotlib_formats[1.2]
rotlib_formats[1.4] = *rotlib_formats[1.3], 'backbone_atoms', 'aln_atoms'
rotlib_formats[1.5] = rotlib_formats[1.4]




from typing import Union, Tuple, List, Set
from pathlib import Path

import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial.distance import cdist
import igraph as ig

from .globals import atom_order, nataa_codes, natnu_codes, mm_backbones
from .Topology import get_min_topol, guess_bonds, modified_bfs_edges


def sort_pdb(pdbfile: Union[str, List[str], List[List[str]]],
             uniform_topology: bool = True,
             index: bool = False,
             bonds: Union[ArrayLike, Set] = set(),
             **kwargs) -> Union[List[str], List[List[str]], List[int]]:
    """Read ATOM lines of a pdb and sort the atoms according to chain, residue index, backbone atoms and side chain atoms.
    Side chain atoms are sorted by distance to each other/backbone atoms with atoms closest to the backbone coming
    first and atoms furthest from the backbone coming last. This sorting is essential to making internal-coordinates
    with consistent and preferred dihedral definitions.

    Parameters
    ----------
    pdbfile : str, List[str], List[List[str]]
        Name of the PDB file, a list of strings containing ATOM lines of a PDB file or a list of lists containing
        ATOM lines of a PDB file, where each sublist corresponds to a state/model of a multi-state pdb.
    uniform_topology: bool
        When given a multi-state pdb, assume that all states have the same topology (bonds) as the first state.
    index: bool :
         Return the sorted index rather than the sorted lines.
    bonds: ArrayLike :
         When sorting the PDB, use the provided bond list to as the topology rather than guessing the bonds.
    kwargs : dict
        Additional keyword arguments.

        aln_atoms : ArrayLike
            Only used with create_library. Array of 3 atom names indicating which atoms will be the alignment atoms for
            the library. These atoms will be used to determine which atoms constitute the backbone and side chain atoms.
    Returns
    -------
    lines : List[str], List[List[str]]
        Sorted list of strings corresponding to the ATOM entries of a PDB file.
    """
    if isinstance(pdbfile, (str, Path)):
        with open(pdbfile, "r") as f:
            lines = f.readlines()

        start_idxs = []
        end_idxs = []
        connect = []
        lines = [line for line in lines if line.startswith(('MODEL', 'ENDMDL', 'CONECT', 'ATOM', 'HETATM'))]
        for i, line in enumerate(lines):
            if line.startswith('MODEL'):
                start_idxs.append(i + 1)
            elif line.startswith("ENDMDL"):
                end_idxs.append(i)
            elif line.startswith('CONECT'):
                connect.append(line)

        # Use connect information for bonds if present
        if connect != [] and bonds == set():
            connect, _, _ = parse_connect(connect)
            kwargs['additional_bonds'] = kwargs.get('additional_bonds', set()) | connect

        # If it's a multi-state pdb...
        if start_idxs != []:

            if uniform_topology:
                # Assume that all states have the same topology as the first
                idxs = _sort_pdb_lines(lines[start_idxs[0]:end_idxs[0]], bonds=bonds, index=True, **kwargs)

            else:
                # Calculate the shared topology and force it
                atom_lines = [lines[s:e] for s, e in zip(start_idxs, end_idxs)]
                min_bonds_list = get_min_topol(atom_lines, forced_bonds=bonds)
                idxs = _sort_pdb_lines(lines[start_idxs[0]:end_idxs[0]], bonds=min_bonds_list, index=True, **kwargs)

            if isinstance(idxs, tuple):
                idxs, bonds = idxs

            lines[:] = [[lines[idx + start][:6] + f"{i + 1:5d}" + lines[idx + start][11:]
                         for i, idx in enumerate(idxs)]
                        for start in start_idxs]

            if kwargs.get('return_bonds', False):
                lines = lines, bonds
        else:
            lines = _sort_pdb_lines(lines, bonds=bonds, index=index, **kwargs)

    elif isinstance(pdbfile, list):
        lines = _sort_pdb_lines(pdbfile, bonds=bonds, index=index, **kwargs)

    return lines


def _sort_pdb_lines(lines, bonds=None, index=False, **kwargs) -> \
        Union[List[str], List[int], Tuple[list[str], List[Tuple[int]]]]:
    """
    Helper function to sort PDB ATOM and HETATM lines based off of the topology of the topology of the molecule.

    Parameters
    ----------
    lines : List[str]
        A list of the PDB ATOM and HETATM lines.
    bonds : Set[Tuple[int]]
        A Set of tuples of atom indices corresponding to atoms ( `lines` ) that are bound to each other. Passing this
        keyword argument will prevent chiLife from guessing the bonds. To guess most bonds but strictly enforce a few
        use he `additional_bonds` keyword argument.
    index : bool
        If True a list of atom indices will be returned
    **kwargs : dict
        Additional keyword arguments.
        return_bonds : bool
            Return bond indices as well, usually only used when letting the function guess the bonds.
        additional_bonds: set(tuple(int))
            Additional bonds (index pairs) that will be added to the bonds list. Used when letting the function guess
            the bonds, but
        aln_atoms : ArrayLike
            Only used with create_library. Array of 3 atom names indicating which atoms will be the alignment atoms for
            the library. These atoms will be used to determine which atoms constitute the backbone and side chain atoms.

    Returns
    -------
    lines : List[str | int]
        The sorted lines or indices corresponding to the sorted lines.
    bonds: Set[Tuple[int]]
        A set of tuples containing pars of indices corresponding to the atoms bound to in lines.
    """

    waters = [line for line in lines if line[17:20] in ('SOL', 'HOH')]
    water_idx = [idx for idx, line in enumerate(lines) if line[17:20] in ('SOL', 'HOH')]
    lines = [line for line in lines if line.startswith(("ATOM", "HETATM")) and line[17:20] not in ('SOL', 'HOH')]
    n_atoms = len(lines)
    index_key = {line[6:11]: i for i, line in enumerate(lines)}

    # Presort
    lines.sort(key=atom_sort_key)
    presort_idx_key = {line[6:11]: i for i, line in enumerate(lines)}
    presort_bond_key = {index_key[line[6:11]]: i for i, line in enumerate(lines)}

    coords = np.array([[float(line[30:38]), float(line[38:46]), float(line[46:54])] for line in lines])
    atypes = np.array([line[76:78].strip() for line in lines])
    anames = np.array([line[12:17].strip() for line in lines])

    if bonds:
        input_bonds = {tuple(b) for b in bonds}
        presort_bonds = set(tuple(sorted((presort_bond_key[b1], presort_bond_key[b2]))) for b1, b2 in bonds)
    else:
        bonds = guess_bonds(coords, atypes)
        presort_bonds = set(tuple(sorted((b1, b2))) for b1, b2 in bonds)
        if kwargs.get('additional_bonds', set()) != set():
            presort_bonds.union(kwargs['additional_bonds'])

    # get residue groups
    chain, resi, resn = lines[0][21], int(lines[0][22:26].strip()), lines[0][17:20].strip()
    start = 0
    resdict = {}
    for curr, pdb_line in enumerate(lines):

        if chain != pdb_line[21] or resi != int(pdb_line[22:26].strip()):
            resdict[chain, resi] = start, curr, resn
            start = curr
            chain, resi, resn = pdb_line[21], int(pdb_line[22:26].strip()), pdb_line[17:20].strip()

    resdict[chain, resi] = start, curr + 1, resn
    midsort_key = []
    for key in resdict:
        start, stop, resn = resdict[key]
        midsort_key += sort_residue(coords, atypes[start:stop], anames[start:stop],
                                    resn, presort_bonds, start, kwargs.get('aln_atoms', None))

    lines[:] = [lines[i] for i in midsort_key]
    lines.sort(key=atom_sort_key)

    if 'input_bonds' not in locals():
        input_bonds = presort_bonds
        idxmap = {presort_idx_key[line[6:11]]: i for i, line in enumerate(lines)}
    else:
        idxmap = {index_key[line[6:11]]: i for i, line in enumerate(lines)}

    # Return line indices if requested
    if index:
        str_lines = lines
        lines = [index_key[line[6:11]] for line in lines] + water_idx

    # Otherwise make new indices
    else:
        lines = [line[:6] + f"{i + 1:5d}" + line[11:] for i, line in enumerate(lines)] + waters

    if kwargs.get('return_bonds', False):
        bonds = {tuple(sorted((idxmap[a], idxmap[b]))) for a, b in input_bonds}
        return lines, bonds

    return lines


def sort_residue(coords, atypes, anames, resname, presort_bonds, start, aln_atoms=None):
    """
    Sort atoms of a residue.

    Parameters
    ----------
    coords : ArrayLike
        cartesian coordinates of each atom.
    atypes : ArrayLike
        Element symbols of each atom.
    anames : ArrayLike
        names of each atom
    resname : str
        Name of the residue
    presort_bonds : ArrayLike
        Array of index pairs corresponding to bonded atom pairs. Index is based on the indices of atoms before sorting.
    start : int
        The index of the first atom of the residue in the greater protein environment.
    aln_atoms : List
        List of atom indices defining the "origin" of the residue. These will be used to define backbone and side chain
        atoms.

    Returns
    -------
    sorted_args : List[int]
        Indices of the residue atoms based on the topology and the backbone/aln atoms.
    """

    stop = start + len(anames)
    n_heavy = np.sum(atypes != 'H')
    res_coords = coords[start:stop]

    # Adjust bond atom indices to reference local index
    bonds = np.array([(a - start, b - start) for a, b in presort_bonds
                      if (start <= a < stop) and (start <= b < stop)])

    # Get all nearest neighbors and sort by distance for generating graph
    distances = np.linalg.norm(res_coords[bonds[:, 0]] - res_coords[bonds[:, 1]], axis=1)
    distances = np.around(distances, decimals=3)

    # Sort bonds by distance
    idx_sort = np.lexsort((bonds[:, 0], bonds[:, 1], distances))
    pairs = bonds[idx_sort]

    # Create graph
    graph = ig.Graph(edges=pairs)

    # Determine if this residue fits aln_atoms criteria. aln_atoms must be defined, present, and adjacent in the graph
    if aln_atoms is not None:
        if len(aln_atoms) != 3:
            raise ValueError('aln_atoms must have length 3')

        aln_cen = np.argwhere(aln_atoms[1] == anames).flat[0]
        neighbor_idx = np.argwhere(np.isin(anames, aln_atoms[::2])).flatten()
        aln_present = np.all(np.isin(aln_atoms, anames))
        aln_adjacent = np.all(np.isin(neighbor_idx, graph.neighbors(aln_cen)))
        use_aln = aln_adjacent and aln_present
    else:
        neighbor_idx = []
        use_aln = False

    # let downstream functions decided backbone
    if use_aln:
        sorted_args = [np.argwhere(anames == a).flat[0] for a in aln_atoms]
        root_idx = np.argwhere(anames == aln_atoms[1]).flat[0]

    # Otherwise attempt to find canonical backbone definitions
    else:
        # guess possible backbone atoms for the residue type with the given atom names
        bb_candidates = get_bb_candidates(anames, resname)

        # Use candidates if they are defined
        if bb_candidates:
            aname_lst = anames.tolist()
            sorted_args = [aname_lst.index(name) for name in bb_candidates if name in aname_lst]

        # Otherwise, if there are no preceding residues start from the first atom
        elif start == 0:
            sorted_args = [0]

        # Otherwise use the atom bound to the nearest residue.
        else:
            for a, b in presort_bonds:
                # Find the atom bonded to a previous residue
                if a < start and start <= b < stop and atypes[b-start] != 'H':
                    sorted_args = [b - start]
                    break
            # Otherwise get the closest to any previous atom
            else:
                dist_mat = cdist(coords[:start], coords[start:stop])
                sorted_args = [np.squeeze(np.argwhere(dist_mat == dist_mat.min()))[0]]

        # If using a defined backbone, use the known root atoms for the BFS search below
        bb_names = anames[sorted_args]

        # TODO: Need a better way to guess the root atom from the backbone and side chains.
        for name in ["CA", "C1'"]:
            if name in bb_names:
                root_idx = np.argwhere(bb_names == name).flat[0]
                root_idx = sorted_args[root_idx]
                break
        # Unless there are no known root atoms, then just use the first atom
        else:
            root_idx = sorted_args[0]

    # Skip BFS sorting if heavy atoms are sorted, i.e. the backbone is known and there is no side chain (e.g. GLY)
    if len(sorted_args) != n_heavy:

        graph = ig.Graph(edges=pairs)

        if root_idx not in graph.vs.indices:
            raise RuntimeError(f'The connectivity of one or more residue is not valid')

        # Use dfs verts for backbone
        if use_aln:
            sorted_args += get_backbone_atoms(graph, root_idx, neighbor_idx, sorted_args=sorted_args)

        # Use bfs verts for
        # bfsv, si, bfse = graph.bfs(root_idx)
        # [v for v in bfsv if v not in sorted_args]
        sidechain_nodes = [v[1] for v in modified_bfs_edges(pairs, root_idx, sorted_args) if v[1] not in sorted_args]

        # Check for disconnected parts of residue
        if not graph.is_connected():
            for g in graph.connected_components():
                if np.any([arg in g for arg in sorted_args]):
                    continue
                free_nodes = [idx for idx in sidechain_nodes if atypes[start + idx] != 'H']
                g_nodes = [idx for idx in g if atypes[start + idx] != 'H']
                near_root = cdist(res_coords[free_nodes], res_coords[g_nodes]).argmin()

                yidx = near_root % len(g_nodes)
                subnodes, _, _ = graph.bfs(g_nodes[yidx])
                sidechain_nodes += list(subnodes)

    elif stop - start > n_heavy:
        # Assumes  non-heavy atoms come after the heavy atoms, which should be true because of the pre-sort
        sidechain_nodes = list(range(n_heavy, n_heavy + (stop - start - len(sorted_args))))
    else:
        sidechain_nodes = []

    sorted_args += sidechain_nodes

    # get any leftover atoms (eg HN)
    if len(sorted_args) != stop - start:
        for idx in range(stop - start):
            if idx not in sorted_args:
                sorted_args.append(idx)

    return [x + start for x in sorted_args]


def get_bb_candidates(anames, resname):
    """
    Function to guess the backbone atom names for a set of atom names residue name.
    Parameters
    ----------
    anames : ArrayLike
        Array of atom names of the residue.
    resname : str
        Residue name

    Returns
    -------
    bb_candidates : ArrayLike
        Array of backbone atom names.

    """
    if resname in natnu_codes:
        bb_candidates = mm_backbones['nu']
    elif resname in nataa_codes:
        bb_candidates = mm_backbones['aa']
    elif np.isin(mm_backbones['aa'], anames).sum() >= 3:
        bb_candidates = mm_backbones['aa']
    elif np.isin(mm_backbones['nu'], anames).sum() >= 3:
        bb_candidates = mm_backbones['nu']
    elif resname in mm_backbones:
        bb_candidates = mm_backbones[resname]
    else:
        bb_candidates = []

    return bb_candidates


def get_backbone_atoms(graph, root_idx, neighbor_idx, **kwargs):
    """
    Function to get backbone atom indices from a graph using a depth first search (dfs), given a root index and a list
    of neighboring indices that define  the starting branches of the backbone atoms (i.e. not sidechain atoms).

    Parameters
    ----------
    graph: igraph.Graph
        The graph to search for backbone atoms.
    root_idx: int
        The root index of the graph that separates the side chain and backbone atoms.
    neighbor_idx: indices of the neighbors of the `root_idx` that start the dfs of the backbone atoms.

    kwargs: dict
        Additional keyword arguments to pass to the depth first search.

        sorted_args: List[int]
            A sorted list of atom indices dictating backbone atoms that have already been defined

    Returns
    -------
    backbone_vs: List[int]
        list of sorted backbone atom indices.

    """

    sorted_args = kwargs.get('sorted_args', [])

    # Start stemming from root atom
    neighbors = graph.neighbors(root_idx)
    backbone_vs = []
    sidechain_vs = []
    bblist_order = {}
    dfsv, dfse = graph.dfs(root_idx)
    for idx in dfsv:
        if idx == root_idx:
            continue
        elif idx in neighbor_idx:
            bblist_order[idx] = len(backbone_vs)
            active_list = backbone_vs
        elif idx in neighbors:

            active_list = sidechain_vs

        if idx not in sorted_args:
            active_list.append(idx)

    bblist_order['end'] = len(backbone_vs)
    bbidx_slices = {}
    items = iter(bblist_order)
    k1 = next(items)
    for k2 in items:
        bbidx_slices[k1] = bblist_order[k1], bblist_order[k2]
        k1 = k2
    a, b = neighbor_idx
    return backbone_vs[slice(*bbidx_slices[b])] + backbone_vs[slice(*bbidx_slices[a])]


def atom_sort_key(pdb_line: str) -> Tuple[str, int, int]:
    """Assign a base rank to sort atoms of a pdb.

    Parameters
    ----------
    pdb_line : str
        ATOM line from a pdb file as a string.

    Returns
    -------
    tuple :
        chain_id, resid, name_order.
        ordered ranking of atom for sorting the pdb.
    """
    chain_id = pdb_line[21]
    res_name = pdb_line[17:20].strip()
    resid = int(pdb_line[22:26].strip())
    atom_name = pdb_line[12:17].strip()
    atom_type = pdb_line[76:79].strip()
    if res_name == "ACE":
        if atom_type != 'H' and atom_name not in ('CH3', 'C', 'O'):
            raise ValueError(f'"{atom_name}" is not canonical name of an ACE residue atom. \n'
                             f'Please rename to "CH3", "C", or "O"')
        name_order = (
            {"CH3": 0, "C": 1, "O": 2}.get(atom_name, 4) if atom_type != "H" else 5
        )

    else:
        name_order = atom_order.get(atom_name, 4) if atom_type != "H" else atom_order.get(atom_name, 7)

    return chain_id, resid, name_order


def parse_connect(connect: List[str]) -> Tuple[Set[Tuple[int]]]:
    """
    Parse PDB CONECT information to get a list covalent bonds, hydrogen bonds and ionic bonds.

    Parameters
    ----------
    connect : List[str]
        A list of strings that are the CONECT lines from a PDB file.

    Returns
    -------
    c_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are bound covalently.
    h_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are hydrogen bound.
    ci_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are bound ionically.
    """
    c_bonds, h_bonds, i_bonds = set(), set(), set()
    for line in connect:
        line = line.ljust(61)
        a0 = int(line[6:11])
        c_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in line[11:31].split()}
        h_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in (line[31:41].split() + line[46:56].split())}
        i_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in (line[41:46], line[56:61]) if not b.isspace()}

    return c_bonds, h_bonds, i_bonds
